"""
    Optimizer.OptStrategyDialog.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy.Trimming import seem_to_be_manually_trimmed
from molass_legacy.KekLib.BraceWidgets import ClosingBrace
from .OptConstants import MIN_NUM_PEAKS, MAX_NUM_PEAKS
from .DefaultNumPeaks import get_default_num_peaks
from molass_legacy.Baseline.BaselineUtils import get_default_baseline_type
from .ElutionComposer import COMPOSER_CB_TEXT
from molass_legacy._MOLASS.Version import is_developing_version

BETA_RELEASE = get_setting("beta_release")
ENABLE_SEPARATE_FOULING = get_setting("enable_separate_fouling")
SHOW_RG_CURVE_ENTRY = False
SHOW_OTHER_OPTIONS = False
SIMPLE_EOII_OPTIONS = True
SHOW_FULL_OPTIONS = False
ENABLE_ADVANCED_SETTINGS = False
SHOW_OPTIMIZATION_STRATEGY_OPTION = False

DEFAULT_FUNC_ITEM = {
    0 : 'default_func_egh',
    1 : 'default_func_sdm',
    2 : 'default_func_lj_egh',
    3 : 'default_func_fd_emg',
    4 : 'default_func_rt_egh',
    5 : 'default_func_edm',
    }

MODEL_LIST = [  (0, "[%d] %s - EGH - Exponential-Gaussian Hybrid"),
                (2, "[%d] %s - Constrained EGH (Plate Theory)"),
                (3, "[%d] %s - Constrained EMG (Plate Theory)"),
                # (4, "[%d] %s - Constrained EMG (Approximation from Rate Theory Moments)"),
                (1, "[%d] %s - SDM - Stochastic DIspersive Model"),
                (5, "[%d] %s - EDM - Equilibrium Dispersive Model"),
                ]

MODEL_TO_FUNC = {
    0 : 'G0346',
    1 : 'G1100',
    5 : 'G2010',
}

class OptStrategyDialog(Dialog):
    def __init__(self, parent, sd, pre_recog=None, test_mode=False):
        self.parent = parent
        self.sd = sd
        self.pre_recog = pre_recog
        self.manually_trimmed = seem_to_be_manually_trimmed()
        self.default_num_peaks = get_default_num_peaks(sd)
        self.major_peak_index = None    # let the user give this info because there is no reliable index yet
        self.applied = False
        location = None if test_mode else "lower center"
        Dialog.__init__(self, parent, "Optimization Strategy", visible=False, location=location)

    def show(self):
        self._show()

    def body(self, body_frame, devel=True):
        if devel:
            from importlib import reload
            import molass_legacy.GuiParts.ColumnTypeSelector
            reload(molass_legacy.GuiParts.ColumnTypeSelector)
        from molass_legacy.Experiment.DataUtils import get_columntype
        from molass_legacy.GuiParts.ColumnTypeSelector import ColumnTypeSelector

        HasA_debug = False

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=50, pady=10)

        indent_width = 50

        # Exclusion Limit
        grid_row = 0
        label = Tk.Label(iframe, text="Exclusion Limit (which implies the SEC pore size)")
        label.grid(row=grid_row, column=0, sticky=Tk.W)

        grid_row += 1
        option_frame = Tk.Frame(iframe)
        option_frame.grid(row=grid_row, column=0, padx=indent_width, pady=5, sticky=Tk.W)

        columntype = get_columntype()
        self.default_excl_limit = columntype.excl_limit
        column_name = columntype.name
        numplates_pm = columntype.num_plates
        if numplates_pm is None:
            numplates_pm = round(get_setting('num_plates_pc')/0.3)

        self.el_option = Tk.IntVar()
        if column_name == 'Unknown':
            el_option = 1
        else:
            el_option = 0
        self.el_option.set(el_option)
        self.excl_limit = Tk.DoubleVar()
        self.excl_limit.set(self.default_excl_limit)
        self.column_name = Tk.StringVar()

        for k, cname in enumerate([ "use the default column value, i.e., %s (%g kDa)" % (column_name, self.default_excl_limit),
                                    "select from the column products list",
                                    "use the specified limit in the next entry: "]):
            button_frame = Tk.Frame(option_frame)
            rb = Tk.Radiobutton(button_frame, text=cname,
                        variable=self.el_option, value=k,
                        )
            rb.pack(side=Tk.LEFT)
            if k == 1:
                self.column_selector = ColumnTypeSelector(button_frame,
                                                          column_name_var=self.column_name,
                                                          excl_limit_var=self.excl_limit)
                self.column_selector.pack(side=Tk.LEFT)
            elif k == 2:
                self.el_entry  = Tk.Entry(button_frame, textvariable=self.excl_limit,
                                  justify=Tk.CENTER, width=6)
                self.el_entry.pack(side=Tk.LEFT)
                unit_label = Tk.Label(button_frame, text=" (kDa)")
                unit_label.pack(side=Tk.LEFT)
                space = Tk.Frame(button_frame, width=60)
                space.pack(side=Tk.LEFT)

            button_frame.grid(row=k, column=1, sticky=Tk.W )

        # Selected Column Display
        selected_column_frame = Tk.Frame(option_frame)
        selected_column_frame.grid(row=0, column=2, rowspan=3)

        brace = ClosingBrace(selected_column_frame, width=20, height=75)
        brace.grid(row=0, column=0)

        excl_limit_disp = Tk.Label(selected_column_frame, textvariable=self.excl_limit)
        excl_limit_disp.grid(row=0, column=1)
        label = Tk.Label(selected_column_frame, text="kDa")
        label.grid(row=0, column=2)
        mwrg_fig_button = Tk.Button(selected_column_frame, text="Show MwRg figure", command=self.show_mwrg_figure)
        mwrg_fig_button.grid(row=0, column=3, padx=5)

        # Tracers
        self.el_option.trace_add("write", self.el_option_tracer)
        self.el_option_tracer()
        self.excl_limit.trace_add("write", self.excl_limit_tracer)

        # Rg Curve Input
        if SHOW_RG_CURVE_ENTRY:
            grid_row += 1
            label = Tk.Label(iframe, text="Rg Curve Input")
            label.grid(row=grid_row, column=0, sticky=Tk.W)

            grid_row += 1
            button_frame = Tk.Frame(iframe)
            button_frame.grid(row=grid_row, column=0, padx=indent_width, pady=10, sticky=Tk.W)

            rg_curve_folder = get_setting("rg_curve_folder")
            self.rg_curve_input = Tk.IntVar()
            self.rg_curve_input.set(0 if rg_curve_folder is None else 1)
            btn_row = 0
            for k, cname in enumerate([ "calculate",
                                        "copy from the Rg curve folder below"]):
                rb = Tk.Radiobutton(button_frame, text=cname,
                            variable=self.rg_curve_input, value=k,
                            )
                rb.grid(row=k, column=0, sticky=Tk.W)

            grid_row += 1
            self.rg_curve_folder = Tk.StringVar()
            if rg_curve_folder is not None:
                self.rg_curve_folder.set(rg_curve_folder)
            entry_frame = Tk.Frame(iframe)
            entry_frame.grid(row=grid_row, column=0)
            space = Tk.Frame(entry_frame, width=indent_width + 30)
            space.grid(row=0, column=0)
            self.rg_folder_entry = FolderEntry(entry_frame, textvariable=self.rg_curve_folder, width=70,
                                                slimbutton=True,
                                                on_entry_cb=self.on_entry_rg_folder )
            self.rg_folder_entry.grid(row=0, column=1, sticky=Tk.W)

            space = Tk.Frame(entry_frame, height=10)
            space.grid(row=1, column=0)
        else:
            rg_curve_folder = get_setting("rg_curve_folder")
            self.rg_curve_input = Tk.IntVar()
            self.rg_curve_input.set(0 if rg_curve_folder is None else 1)
            self.rg_curve_folder = Tk.StringVar()
            if rg_curve_folder is not None:
                self.rg_curve_folder.set(rg_curve_folder)

        # Number of Components
        grid_row += 1
        label = Tk.Label(iframe, text="Number of Components")
        label.grid(row=grid_row, column=0, sticky=Tk.W)

        grid_row += 1
        option_frame = Tk.Frame(iframe)
        option_frame.grid(row=grid_row, column=0, padx=indent_width, pady=5, sticky=Tk.W)

        self.nc_option = Tk.IntVar()
        self.nc_option.set(0)
        self.num_peaks = Tk.IntVar()
        set_num_peaks = self.default_num_peaks
        self.num_peaks.set(set_num_peaks)

        btn_row = 0
        for k, cname in enumerate([ "use an automatically determined number",
                                    "use the specified number in the next box: "]):
            button_frame = Tk.Frame(option_frame)
            rb = Tk.Radiobutton(button_frame, text=cname,
                        variable=self.nc_option, value=k,
                        )
            rb.pack(side=Tk.LEFT)
            if k == 1:
                self.sbox  = Tk.Spinbox(button_frame, textvariable=self.num_peaks,
                                  from_=MIN_NUM_PEAKS, to=MAX_NUM_PEAKS, increment=1,
                                  justify=Tk.CENTER, width=6)
                self.sbox.pack(side=Tk.LEFT)

            button_frame.grid(row=btn_row+k, column=1, sticky=Tk.W )

        self.nc_option.trace_add("write", self.nc_option_tracer)
        self.nc_option_tracer()
        self.num_peaks.trace_add("write", self.num_peaks_tracer)

        # Elution Model
        grid_row += 1
        label = Tk.Label(iframe, text="Elution Model")
        label.grid(row=grid_row, column=0, sticky=Tk.W)

        grid_row += 1
        option_frame = Tk.Frame(iframe)
        option_frame.grid(row=grid_row, column=0, padx=indent_width, pady=5, sticky=Tk.W)

        elution_model = get_setting("elution_model")

        self.elution_model_gui = Tk.IntVar()
        self.elution_model_gui.set(elution_model)

        self.number_of_plates = Tk.DoubleVar()
        self.number_of_plates.set(numplates_pm)

        i = 0
        if SHOW_FULL_OPTIONS:
            model_list = MODEL_LIST
        else:
            model_list = [MODEL_LIST[0], MODEL_LIST[3], MODEL_LIST[4]]
        for k, cname_fmt in model_list:
            state = Tk.DISABLED if k == 4 else Tk.NORMAL
            # state = Tk.NORMAL
            cname = cname_fmt % (k, get_setting(DEFAULT_FUNC_ITEM[k]))
            rb = Tk.Radiobutton(option_frame, text=cname, variable=self.elution_model_gui, value=k, state=state)
            span = 3 if k == 4 else 2
            rb.grid(row=i, column=0, columnspan=span, sticky=Tk.W )

            i += 1

        if SHOW_FULL_OPTIONS:
            # Palte Number
            space = Tk.Frame(option_frame, width=20)
            space.grid(row=1, column=2)
            brace = ClosingBrace(option_frame, width=20, height=50)
            brace.grid(row=1, column=3, rowspan=2)

            pn_frame = Tk.Frame(option_frame, width=20)
            pn_frame.grid(row=1, column=4, rowspan=2)
            label = Tk.Label(pn_frame, text="Number of Plates: ")
            label.grid(row=0, column=0)
            self.num_plates_entry = Tk.Entry(pn_frame, textvariable=self.number_of_plates, width=8, justify=Tk.CENTER)
            self.num_plates_entry.grid(row=0, column=1)
            percol = Tk.Label(pn_frame, text="/m")
            percol.grid(row=0, column=2)

        # Trimming Strategy
        grid_row += 1
        label = Tk.Label(iframe, text="Trimming Strategy")
        label.grid(row=grid_row, column=0, sticky=Tk.W)

        grid_row += 1
        trimming_frame = Tk.Frame(iframe)
        trimming_frame.grid(row=grid_row, column=0, columnspan=2, padx=indent_width, pady=5, sticky=Tk.W)

        self.trimming_strategy = Tk.IntVar()
        self.trimming_strategy.set(1 if self.manually_trimmed else 2)

        for k, (strategy_text, v) in enumerate([
                                        ("apply automatic trimming", 2),
                                        ("apply traditional (or manually specified) trimming", 1),
                                        ]):
            rb = Tk.Radiobutton(trimming_frame, text=strategy_text,
                        variable=self.trimming_strategy, value=v,
                        )
            rb.grid(row=k, column=0, sticky=Tk.W )

        self.trimming_strategy.trace_add("write", self.trimming_strategy_tracer)

        state = Tk.DISABLED if BETA_RELEASE else Tk.NORMAL

        # Trimming Check Figure
        space = Tk.Frame(trimming_frame, width=10)
        space.grid(row=0, column=1)
        check_fig_frame = Tk.Frame(trimming_frame)
        check_fig_frame.grid(row=0, column=2, rowspan=2)
        brace = ClosingBrace(check_fig_frame, width=20, height=50)
        brace.grid(row=0, column=0)
        button  = Tk.Button(check_fig_frame, text="Check Figure", command=self.show_trimming_figure)
        button.grid(row=0, column=1, padx=50)

        # Option Grid Frame
        grid_row += 1
        grid_frame = Tk.Frame(iframe)
        grid_frame.grid(row=grid_row, column=0, padx=0, pady=5, sticky=Tk.W)
        for k in range(5):
            grid_frame.columnconfigure(k, weight=1)

        # Baseline Options
        gf_row = 0
        label = Tk.Label(grid_frame, text="Baseline Options")
        label.grid(row=gf_row, column=0, sticky=Tk.W)

        gf_row += 1
        space = Tk.Frame(grid_frame)
        space.grid(row=gf_row, column=0)

        self.unified_baseline_type = Tk.IntVar()
        baseline_type = get_default_baseline_type()
        # baseline_type = get_setting("unified_baseline_type")
        self.unified_baseline_type.set(baseline_type)

        if ENABLE_SEPARATE_FOULING:
            option_specs =  [  # (0, "No baseline separation"),
                            (1, "Linear"),
                            (2, "Linear + Uniform Fouling"),
                            (3, "Linear + Separate Fouling"),
                            ]
        else:
            option_specs = [  # (0, "No baseline separation"),
                            (1, "Linear"), 
                            (2, "Linear + Fouling")
                            ]
            
        for k, cname in (option_specs):
            state = Tk.DISABLED if k == 3 else Tk.NORMAL
            # option_frame.columnconfigure(k, weight=1)
            rb = Tk.Radiobutton(grid_frame, text=cname, variable=self.unified_baseline_type, value=k, state=state)
            rb.grid(row=gf_row, column=k, sticky=Tk.W )

        self.optimization_method = Tk.IntVar()
        self.optimization_method.set(get_setting("optimization_method"))

        gf_row += 1
        label = Tk.Label(grid_frame, text="Optimization Method")
        label.grid(row=gf_row, column=0, sticky=Tk.W)

        gf_row += 1
        option_specs = [
                        (0, "Basin-Hopping"),
                        (1, "Nested Sampling"),
                        (2, "Alternate BH → NS"),
                        (3, "Alternate NS → BH"),
                        # (2, "MCMC (emcee)        "),
                        # (3, "SMC (pyABC)"),
                        ]
        # activate = slice(None, None) if is_developing_version() else slice(0,2)
        activate = slice(0,2)
        for k, cname in (option_specs[activate]):
            # method_frame.columnconfigure(k, weight=1)
            rb = Tk.Radiobutton(grid_frame, text=cname, variable=self.optimization_method, value=k)
            rb.grid(row=gf_row, column=k+1, sticky=Tk.W)

        if SHOW_OPTIMIZATION_STRATEGY_OPTION:
            gf_row += 1
            label = Tk.Label(grid_frame, text="Optimization Strategy")
            label.grid(row=gf_row, column=0, sticky=Tk.W)

            self.optimization_strategy = Tk.IntVar()
            self.optimization_strategy.set(get_setting("optimization_strategy"))

            gf_row += 1

            option_specs = [
                            (0, "Standard     "),
                            (1, "Custom       "),
                            ]
            for k, cname in (option_specs[activate]):
                # method_frame.columnconfigure(k, weight=1)
                rb = Tk.Radiobutton(grid_frame, text=cname, variable=self.optimization_strategy, value=k)
                rb.grid(row=gf_row, column=k+1, sticky=Tk.W)

            self.editor_button = Tk.Button(grid_frame, text="Strategy Editor", command=self.show_strategy_editor, state=Tk.DISABLED)
            self.editor_button.grid(row=gf_row, column=3, sticky=Tk.W)

            self.optimization_strategy.trace_add("write", self.optimization_strategy_tracer)

        # option variables
        self.ratio_interpretation = Tk.IntVar()
        self.ratio_interpretation.set(get_setting("ratio_interpretation"))
        self.try_model_composing = Tk.IntVar()
        self.try_model_composing.set(get_setting('try_model_composing'))
        self.identification_allowance = Tk.DoubleVar()
        self.identification_allowance.set(get_setting("identification_allowance"))
        self.separate_eoii = Tk.IntVar()
        self.separate_eoii.set(get_setting("separate_eoii"))
        self.separate_eoii_type = Tk.IntVar()
        self.apply_sf_bounds = Tk.IntVar()
        self.apply_sf_bounds.set(get_setting("apply_sf_bounds"))
        self.sf_bound_ratio = Tk.DoubleVar()
        self.sf_bound_ratio.set(get_setting("sf_bound_ratio"))
        self.avoid_peak_fronting = Tk.IntVar()
        self.avoid_peak_fronting.set(get_setting("avoid_peak_fronting"))
        self.apply_rg_discreteness = Tk.IntVar()
        self.apply_rg_discreteness.set(get_setting("apply_rg_discreteness"))
        self.rg_discreteness_unit = Tk.DoubleVar()
        self.rg_discreteness_unit.set(get_setting("rg_discreteness_unit"))
        self.apply_mw_integrity = Tk.IntVar()
        self.apply_mw_integrity.set(get_setting("apply_mw_integrity"))
        self.mw_integer_ratios = Tk.StringVar()
        ratios = get_setting("mw_integer_ratios")
        if ratios is None:
            ratios = ""
        self.mw_integer_ratios.set(ratios)
        self.avoid_peak_fronting_cb = None
        self.try_model_composing_cb = None
        self.separate_eoii_flags = []
        for k in range(self.num_peaks.get()):
            self.separate_eoii_flags.append(Tk.IntVar())

        if ENABLE_ADVANCED_SETTINGS:
            # Advanced Settings
            grid_row += 1
            label = Tk.Label(iframe, text="Advanced Settings")
            label.grid(row=grid_row, column=0, sticky=Tk.W)

            grid_row += 1
            advanced_frame = Tk.Frame(iframe)
            advanced_frame.grid(row=grid_row, column=0, padx=indent_width, pady=5, sticky=Tk.W)

            frame_row = 0
            cb = Tk.Checkbutton(advanced_frame, text="apply ratio interpretation", variable=self.ratio_interpretation)
            cb.grid(row=frame_row, column=0, sticky=Tk.W)

        if SHOW_FULL_OPTIONS:
            frame_row += 1
            compose_frame = Tk.Frame(advanced_frame)
            compose_frame.grid(row=frame_row, column=0, columnspan=5, sticky=Tk.W)
            cb = Tk.Checkbutton(compose_frame, text=COMPOSER_CB_TEXT, variable=self.try_model_composing)
            cb.grid(row=0, column=0, sticky=Tk.W)
            self.try_model_composing_cb = cb
            space = Tk.Frame(compose_frame, width=25)
            space.grid(row=0, column=1)
            state = Tk.DISABLED
            label = Tk.Label(compose_frame, text="identification allowance: ", state=state)
            label.grid(row=0, column=2, sticky=Tk.W)
            self.identification_allowance_label = label
            entry = Tk.Entry(compose_frame, textvariable=self.identification_allowance, width=6, justify=Tk.CENTER, state=state)
            entry.grid(row=0, column=3, sticky=Tk.W)
            self.identification_allowance_entry = entry
            self.try_model_composing.trace_add("write", self.try_model_composing_tracer)

            frame_row += 1
            separate_eoii_frame = Tk.Frame(advanced_frame)
            separate_eoii_frame.grid(row=frame_row, column=0, columnspan=5, sticky=Tk.W)

            if SIMPLE_EOII_OPTIONS:
                cb = Tk.Checkbutton(separate_eoii_frame, text="separate effect of interparticle interactions", variable=self.separate_eoii)
                cb.grid(row=0, column=0, sticky=Tk.W)
                self.separate_eoii.trace_add("write", self.update_separate_eoii_simple)
                frame = Tk.Frame(separate_eoii_frame)
                frame.grid(row=1, column=0,  sticky=Tk.W, padx=40)
                cb = Tk.Checkbutton(frame, text="applying bounds", variable=self.apply_sf_bounds, state=state)
                cb.grid(row=0, column=0,  sticky=Tk.W)
                self.apply_sf_bounds_cb = cb
                label = Tk.Label(frame, text="with parameter: ", state=state)
                label.grid(row=0, column=1)
                entry = Tk.Entry(frame, textvariable=self.sf_bound_ratio, width=6, justify=Tk.CENTER, state=state)
                entry.grid(row=0, column=2, sticky=Tk.W)

                self.sf_bounds_dependents = [cb, label, entry]
                self.update_separate_eoii_simple()

            else:
                cb = Tk.Checkbutton(separate_eoii_frame, text="separate effect of interparticle interactions in either of the following ways", variable=self.separate_eoii)
                cb.grid(row=0, column=0, sticky=Tk.W)
                self.separate_eoii.trace_add("write", self.update_separate_eoii_flags)
                cb = Tk.Checkbutton(separate_eoii_frame, text="applying bounds", variable=self.apply_sf_bounds, state=state)
                cb.grid(row=0, column=1,  sticky=Tk.W)
                self.apply_sf_bounds_cb = cb

                frame_row += 1
                radio_button_frame = Tk.Frame(advanced_frame)
                radio_button_frame.grid(row=frame_row, column=0, columnspan=5, sticky=Tk.W, padx=40)
                self.separate_eoii_rbs = []
                rb = Tk.Radiobutton(radio_button_frame, text="treating the total sum of components as a single cause elution", variable=self.separate_eoii_type, value=1)
                rb.grid(row=0, column=0, sticky=Tk.W)
                self.separate_eoii_rbs.append(rb)
                self.separate_eoii_type.trace_add("write", self.separate_eoii_type_tracer)

                label_flags_frame = Tk.Frame(radio_button_frame)
                label_flags_frame.grid(row=1, column=0, sticky=Tk.W)
                self.label_flags_frame = label_flags_frame
                rb = Tk.Radiobutton(label_flags_frame, text="separately in the order of component peak positions:", variable=self.separate_eoii_type, value=2)
                rb.grid(row=0, column=0)
                self.separate_eoii_rbs.append(rb)

                # construct flags_frame separately to avoid strange behavior in cases with less than three checkbuttons
                flags_frame = Tk.Frame(label_flags_frame)
                flags_frame.grid(row=0, column=1)
                self.flags_frame = flags_frame

                self.separate_eoii_cbs = []         # needed for the destroying
                self.update_separate_eoii_flags()

            frame_row += 1
            cb = Tk.Checkbutton(advanced_frame, text="avoid peak fronting", variable=self.avoid_peak_fronting)
            cb.grid(row=frame_row, column=0, sticky=Tk.W)
            self.avoid_peak_fronting_cb = cb

            frame_row += 1
            cb = Tk.Checkbutton(advanced_frame, text="apply Rg-discreteness unit", variable=self.apply_rg_discreteness)
            cb.grid(row=frame_row, column=0, sticky=Tk.W)
            entry_frame = Tk.Frame(advanced_frame)
            entry_frame.grid(row=frame_row, column=1, sticky=Tk.W)
            self.discreteness_unit_entry = Tk.Entry(entry_frame, textvariable=self.rg_discreteness_unit, width=5, justify=Tk.CENTER)
            self.discreteness_unit_entry.grid(row=0, column=0, sticky=Tk.W)
            unit_label = Tk.Label(entry_frame, text="Å")
            unit_label.grid(row=0, column=1, sticky=Tk.W)
            self.apply_rg_discreteness_tracer()
            self.apply_rg_discreteness.trace_add("write", self.apply_rg_discreteness_tracer)

            frame_row += 1
            cb = Tk.Checkbutton(advanced_frame, text="apply MW integer ratios", variable=self.apply_mw_integrity)
            cb.grid(row=frame_row, column=0, sticky=Tk.W)
            entry_frame = Tk.Frame(advanced_frame)
            entry_frame.grid(row=frame_row, column=1, sticky=Tk.W)
            self.mw_integer_ratios_entry = Tk.Entry(entry_frame, textvariable=self.mw_integer_ratios, width=10, justify=Tk.CENTER)
            self.mw_integer_ratios_entry.grid(row=0, column=0, sticky=Tk.W)
            guide_message = Tk.Label(entry_frame, text="specify the ratios like [2,1] if you are confident, otherwise leave it empty")
            guide_message.grid(row=0, column=1, sticky=Tk.W, padx=2)
            self.apply_mw_integrity_tracer()
            self.apply_mw_integrity.trace_add("write", self.apply_mw_integrity_tracer)

        # Other Options
        self.strict_sec_penalty = Tk.IntVar()
        self.strict_sec_penalty.set(1)
        self.correction = Tk.IntVar()
        self.correction.set(1)
        self.uv_basemodel = Tk.IntVar()
        self.uv_basemodel.set(1)

        if SHOW_OTHER_OPTIONS or SHOW_FULL_OPTIONS:
            grid_row += 1
            label = Tk.Label(iframe, text="Other Options")
            label.grid(row=grid_row, column=0, sticky=Tk.W)

            grid_row += 1
            cb = Tk.Checkbutton( iframe, text="apply strict size exclusion penalty in the optimization",
                                    variable=self.strict_sec_penalty, state=state )
            cb.grid(row=grid_row, column=0, sticky=Tk.W, padx=indent_width)

            grid_row += 1
            cb = Tk.Checkbutton( iframe, text="apply baseline correction to X-ray data before the optimization",
                                    variable=self.correction, state=state )
            cb.grid(row=grid_row, column=0, sticky=Tk.W, padx=indent_width)

            grid_row += 1
            cb = Tk.Checkbutton( iframe, text="use the Advanced UV Baseline Model",
                                    variable=self.uv_basemodel, state=Tk.DISABLED )     # don't change uv_basemodel == 1 until the non-advanced baseline gets completed
            cb.grid(row=grid_row, column=0, sticky=Tk.W, padx=indent_width)
            self.uv_basemodel_cb = cb

        # Tracers which concers multiple parts
        self.elution_model_gui_tracer()
        self.elution_model_gui.trace_add("write", self.elution_model_gui_tracer)

    def get_model_name(self):
        from molass_legacy.Optimizer.OptimizerUtils import get_model_name
        elution_model = self.elution_model_gui.get()
        func = MODEL_TO_FUNC[elution_model]
        return get_model_name(func)

    def on_entry_rg_folder(self):
        pass

    def el_option_tracer(self, *args):
        el_option = self.el_option.get()
        if el_option == 0:
            self.excl_limit.set(self.default_excl_limit)

        def get_state(state):
            return Tk.NORMAL if state else Tk.DISABLED

        self.column_selector.config(state=get_state(el_option == 1))
        self.el_entry.config(state=get_state(el_option == 2))

    def excl_limit_tracer(self, *args):
        el_option = self.el_option.get()
        try:
            self.excl_limit_disp_var.set("%g kDa" % self.excl_limit.get())
        except:
            pass

    def elution_model_gui_tracer(self, *args):
        elution_model_gui = self.elution_model_gui.get()

        if SHOW_FULL_OPTIONS:
            state = Tk.NORMAL if elution_model_gui in [2, 3] else Tk.DISABLED
            for w in [self.num_plates_entry]:
                w.config(state=state)

        if self.avoid_peak_fronting_cb is not None:
            if elution_model_gui in [0, 2, 3]:
                self.avoid_peak_fronting_cb.config(state=Tk.NORMAL)
            else:
                self.avoid_peak_fronting.set(0)
                self.avoid_peak_fronting_cb.config(state=Tk.DISABLED)

        if self.try_model_composing_cb is not None:
            if elution_model_gui in [0, 2]:
                state = Tk.NORMAL
            else:
                state = Tk.DISABLED
                self.try_model_composing.set(0)
            self.try_model_composing_cb.config(state=state)

    def trimming_strategy_tracer(self, *args):
        strategy = self.trimming_strategy.get()
        if self.manually_trimmed and strategy == 2:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox

            MessageBox.showwarning( "Manual Trimming Cleared Warning",
                    "Manual trimming info seems to exist.\n"
                    "Changing to automatic trimming will clear the info.\n"
                    "Revert this option if it is not desired.",
                    parent=self )

        if False:
            if strategy == 2:
                state = Tk.NORMAL
                advanced = 1
            else:
                state = Tk.DISABLED
                advanced = 0
            self.uv_basemodel_cb.config(state=state)
            self.uv_basemodel.set(advanced)

    def nc_option_tracer(self, *args):
        nc_option = self.nc_option.get()
        state = Tk.DISABLED if nc_option == 0 else Tk.NORMAL
        self.sbox.config(state=state)

    def num_peaks_tracer(self, *args):
        if not SIMPLE_EOII_OPTIONS:
            self.update_separate_eoii_flags()

    def try_model_composing_tracer(self, *args):
        try_model_composing = self.try_model_composing.get()
        state = Tk.NORMAL if try_model_composing else Tk.DISABLED
        for w in (  self.identification_allowance_label,
                    self.identification_allowance_entry,
                    ):
            w.config(state=state)

    def update_separate_eoii_simple(self, *args):
        if self.separate_eoii.get():
            state = Tk.NORMAL
            type_value = 1
        else:
            state = Tk.DISABLED
            type_value = 0
        self.separate_eoii_type.set(type_value)
        for w in self.sf_bounds_dependents:
            w.config(state=state)

    def update_separate_eoii_flags(self, *args):
        num_peaks = self.num_peaks.get()
        for cb in self.separate_eoii_cbs:
            cb.destroy()

        if self.separate_eoii.get() == 1:
            state = Tk.NORMAL
            self.separate_eoii_type.set(1)
        else:
            state = Tk.DISABLED
            self.separate_eoii_type.set(0)

        for k, rb in enumerate(self.separate_eoii_rbs):
            if k == 0:
                rb.config(state=state)
            else:
                # not yet supported
                rb.config(state=Tk.DISABLED)

        # self.flags_label.config(state=state)
        self.separate_eoii_flags = []
        self.separate_eoii_cbs = []
        for k in range(num_peaks):
            print([k], k)
            flag_value = Tk.IntVar()
            flag_value.set(1 if k == self.major_peak_index else 0)      # note that self.major_peak_index may be None
            self.separate_eoii_flags.append(flag_value)
            cb = Tk.Checkbutton(self.flags_frame, text="", variable=flag_value, state=Tk.DISABLED)  # not yet supported
            cb.grid(row=0, column=k)
            self.separate_eoii_cbs.append(cb)

    def separate_eoii_type_tracer(self, *args):
        separate_eoii_type = self.separate_eoii_type.get()
        state = Tk.NORMAL if separate_eoii_type > 0 else Tk.DISABLED
        self.apply_sf_bounds_cb.config(state=state)
        if separate_eoii_type < 2:
            return

        state = Tk.NORMAL if separate_eoii_type == 2 else Tk.DISABLED
        for cb in self.separate_eoii_cbs:
            cb.config(state=state)

    def apply_rg_discreteness_tracer(self, *args):
        apply_rg_discreteness = self.apply_rg_discreteness.get()
        state = Tk.DISABLED if apply_rg_discreteness == 0 else Tk.NORMAL
        self.discreteness_unit_entry.config(state=state)

    def apply_mw_integrity_tracer(self, *args):
        apply_mw_integrity = self.apply_mw_integrity.get()
        state = Tk.DISABLED if apply_mw_integrity == 0 else Tk.NORMAL
        self.mw_integer_ratios_entry.config(state=state)

    def optimization_strategy_tracer(self, *args):
        strategy = self.optimization_strategy.get()
        state = Tk.NORMAL if strategy == 1 else Tk.DISABLED
        self.editor_button.config(state=state)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, padx=20)

        for j in range(2):
            box.columnconfigure(j, weight=1)

        w = Tk.Button(box, text="◀ Cancel", width=10, command=self.cancel)
        col = 0
        w.grid(row=0, column=col, pady=10)
        self.cancel_btn = w

        col += 1
        w = Tk.Button(box, text="▶ Proceed", width=12, command=self.ok)
        w.grid(row=0, column=col, pady=10)
        self.proceed_btn = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def validate(self):
        ret = 1
        if self.apply_mw_integrity.get():
            ratios = self.mw_integer_ratios.get()
            if ratios == "":
                pass
            else:
                try:
                    self.mw_integer_ratios_eval = eval(ratios)
                    assert type(self.mw_integer_ratios_eval) is list
                except Exception as exc:
                    import molass_legacy.KekLib.CustomMessageBox as MessageBox
                    MessageBox.showerror( "Value Error", "mw_integer_ratios error:" + str(exc), parent=self)
                    ret = 0
        return ret

    def apply(self):
        from molass_legacy._MOLASS.SerialSettings import set_setting
        from molass_legacy.Global.V2Init import update_sec_settings

        self.applied = True

        set_setting("uv_basemodel", self.uv_basemodel.get())
        elution_model = self.elution_model_gui.get()
        set_setting("elution_model", elution_model)
        item_name = DEFAULT_FUNC_ITEM[elution_model]
        func_code = get_setting(item_name)
        set_setting("default_objective_func", func_code)

        el_option = self.el_option.get()
        column_name = self.column_name.get()
        excl_limit = self.excl_limit.get()
        num_plates_pm = self.number_of_plates.get()
        update_sec_settings(el_option, column_name, excl_limit, num_plates_pm)

        set_setting("unified_baseline_type", self.unified_baseline_type.get())

        # Advanced Settings
        set_setting("ratio_interpretation", self.ratio_interpretation.get())
        set_setting("optimization_method",self.optimization_method.get())
        set_setting("try_model_composing", self.try_model_composing.get())
        set_setting("separate_eoii", self.separate_eoii.get())
        set_setting("separate_eoii_type", self.separate_eoii_type.get())
        separate_eoii_flags = []
        for flag in self.separate_eoii_flags:
            separate_eoii_flags.append(flag.get())
        set_setting("apply_sf_bounds", self.apply_sf_bounds.get())
        set_setting("sf_bound_ratio", self.sf_bound_ratio.get())
        set_setting("separate_eoii_flags", separate_eoii_flags)
        set_setting("avoid_peak_fronting", self.avoid_peak_fronting.get())
        set_setting("apply_rg_discreteness", self.apply_rg_discreteness.get())
        set_setting("rg_discreteness_unit", self.rg_discreteness_unit.get())
        set_setting("apply_mw_integrity", self.apply_mw_integrity.get())
        ratios = self.mw_integer_ratios.get()
        if ratios == "":
            ratios = None
        else:
            ratios = self.mw_integer_ratios_eval   # this value has been validated
        set_setting("mw_integer_ratios", ratios)
        set_setting("identification_allowance", self.identification_allowance.get())

        if self.trimming_strategy.get() == 2:
            set_setting("uv_restrict_list", get_setting("uv_restrict_copy"))
            set_setting("xr_restrict_list", get_setting("xr_restrict_copy"))
            set_setting("manually_trimmed", False)

    def get_num_peaks(self):
        if self.nc_option.get() == 1:
            num_peaks = self.num_peaks.get()
        else:
            # num_peaks = None      # revive this if a better estimation using Rg curve become availabe
            num_peaks = self.num_peaks.get()
        return num_peaks

    def get_options(self):
        strict_sec_penalty = self.strict_sec_penalty.get()
        correction = self.correction.get()
        trimming = self.trimming_strategy.get()
        uv_basemodel = self.uv_basemodel.get()
        unified_baseline_type = self.unified_baseline_type.get()
        return strict_sec_penalty==1, correction, trimming, unified_baseline_type

    def show_mwrg_figure(self, debug=True):
        if debug:
            from importlib import reload
            import SecTheory.MwRgFigure
            reload(SecTheory.MwRgFigure)
        from SecTheory.MwRgFigure import MwRgFigure
        fig = MwRgFigure(self, self.excl_limit.get())
        fig.show()

    def show_trimming_figure(self, debug=True):
        if debug:
            from importlib import reload
            import Trimming.TrimmingResult
            reload(Trimming.TrimmingResult)
        from molass_legacy.Trimming.TrimmingResult import TrimmingResultDialog
        from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition

        if self.pre_recog is None:
            # doing this twice causes an error. why?
            self.pre_recog = PreliminaryRecognition(self.sd)

        dialog = TrimmingResultDialog(self, self.pre_recog)
        dialog.show()

    def show_strategy_editor(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.StrategyEditor
            reload(Optimizer.StrategyEditor)
        from molass_legacy.Optimizer.StrategyEditor import StrategyEditor

        editor = StrategyEditor(self)
        editor.show()


"""
    IcisDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.DENSS.DenssUtils import fit_data
from SvdDenoise import get_denoised_data

def get_primary_scurve(qv, D, E, ecurve, width=11):
    j = ecurve.peak_info[ecurve.primary_peak_no][1]
    hw = width//2
    eslice = slice(j-hw,j+hw+1)
    return np.array([qv] + [np.average(M[:,eslice], axis=1) for M in (D, E)]).T

def get_primary_eslice(ecurve):
    lower, middle, upper = ecurve.peak_info[ecurve.primary_peak_no]
    eslice = slice(lower,upper+1), middle - lower
    return eslice

def solve_Bq(iq, aq, bqinit, cinit):

    def objective(p):
        bq = p[:-1]
        c = p[-1]
        return np.sum((aq*c + bq*c**2 - iq)**2)

    pinit = np.concatenate([bqinit, [cinit]])
    ret = minimize(objective, pinit)
    print("fun init=", objective(pinit), "ret.nit=", ret.nit, "ret.fun=", ret.fun)
    return ret.x[:-1], ret.x[-1]

def get_theretical_C(C_):
    cinit = C_[0,:]

    def objective(c):
        C = np.array([c, c**2])
        return np.linalg.norm(C - C_)

    ret = minimize(objective, cinit)
    c = ret.x
    return np.array([c, c**2])

def demo(in_folder, sd):
    print(in_folder)

    D, E, qv, ecurve = sd.get_xr_data_separate_ly()

    Iq = get_primary_scurve(qv, D, E, ecurve)
    qc, ac, ec, dmax = fit_data(*Iq.T)

    eslice, j = get_primary_eslice(ecurve)
    D_ = get_denoised_data(D[:,eslice], rank=2)
    E_ = E[:,eslice]

    i_smp = sd.xray_index
    c = D_[i_smp,:]
    C = np.array([c, c**2])
    Cinv = np.linalg.pinv(C)
    P = D_ @ Cinv
    Dinv = np.linalg.pinv(D_)
    W = Dinv @ P
    Pe = np.sqrt(np.dot(E_**2, W**2))

    Aq = np.array([qv, P[:,0], Pe[:,0]]).T
    qc_, ac_, ec_, dmax_ = fit_data(*Aq.T)

    n = len(qc) - len(qv)
    diff = np.min(qc[n:] - qv)

    print(len(qv), len(qc), len(qc_), diff)

    Aq_ = np.array([qv, ac_[n:], ec_[n:]]).T
    Bq = np.array([qv, P[:,1], Pe[:,1]]).T

    cinit = c[j]
    bq_, cs = solve_Bq(Iq[:,1], Aq_[:,1], Bq[:,1], cinit)

    P_ = np.array([Aq_[:,1], bq_]).T
    Pinv = np.linalg.pinv(P_)
    C_ = Pinv @ D_
    Ct = get_theretical_C(C_)

    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
        ax1.set_yscale('log')
        ax1.plot(qv, Iq[:,1])
        ax1.plot(qc, ac)
        ax1.plot(qv, Aq[:,1])
        ax1.plot(qv, Aq_[:,1])

        ax2.plot(qv, Bq[:,1])
        ax2.plot(qv, bq_)

        ax3.plot(c)
        # ax3.plot(C_[0,:])
        ax3.plot(Ct[0,:])

        fig.tight_layout()
        plt.show()



"""
    FdEmgParams.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from .LjEghParams import LjEghParams
from .SimpleSecParams import initial_guess
from SecTheory.SecEstimator import NUM_SEC_PARAMS
from SecTheory.FoleyDorseyEmg import convert_to_xr_params_fd, compute_tau

class FdEmgParams(LjEghParams):
    def __init__(self, n_components):
        LjEghParams.__init__(self, n_components)

    def __str__(self):
        return "FdEmgParams(nc=%d)" % (self.n_components - 1)

    def get_estimator(self, editor, debug=True):
        if debug:
            from importlib import reload
            import Estimators.FdEmgEstimator
            reload(Estimators.FdEmgEstimator)
        from Estimators.FdEmgEstimator import FdEmgEstimator
        return FdEmgEstimator(editor)

    def split_params_simple(self, params, convert=True):
        req_params = params[0:-NUM_SEC_PARAMS]
        sec_params = params[-NUM_SEC_PARAMS:]
        Npc, tI = sec_params[[0,2]]
        split_params = self.split_params(self.n_components, req_params)
        if convert:
            xr_params = convert_to_xr_params_fd(split_params[0], tI, Npc)
        else:
            xr_params = split_params[0]
        self.separate_params = [xr_params] + split_params[1:] + [sec_params]
        return self.separate_params

    def split_as_unified_params(self, params, **kwargs):
        return self.split_params_simple(params, **kwargs)

    def get_xr_param_bounds(self, temp_xr_params, tI, Npc):
        xr_h_max = np.max(temp_xr_params[:,0])
        m_ = np.average(temp_xr_params[:,1])
        s_ = (m_ - tI)/np.sqrt(Npc)
        m_allow = np.max(temp_xr_params[:,1])*0.2
        xr_bounds = []
        for h, mu, As in temp_xr_params:        # temp_xr_params is not converted
            xr_bounds.append((0, xr_h_max*2))                       # h
            xr_bounds.append((max(0, mu - m_allow), mu + m_allow))  # m
            xr_bounds.append((s_*0.5, s_*1.5))                      # s
        return xr_bounds

    def get_xr_parameter_names(self):
        nc = self.n_components - 1
        xr_names = []
        for k in range(nc):
            xr_names += ["$h_%d$" % k, r"$\mu_%d$" % k, r"$\sigma_%d$" % k]
        return xr_names

    def get_peak_pos_array_list(self, x_array):
        n = self.n_components - 1
        pos_array_list = []
        tI = x_array[:,-2]
        Npc = x_array[:,-1]
        for k in range(n):
            mu = x_array[:,3*k+1]
            sigma = x_array[:,3*k+2]
            tau = compute_tau(tI, Npc, mu, sigma)    # note that this is a vector computation case
            pos_array_list.append(mu + tau)
        return pos_array_list

    def get_params_sheet(self, parent, params, dsets, optimizer):
        from .FdEmgParamsSheet import FdEmgParamsSheet
        return FdEmgParamsSheet(parent, params, dsets, optimizer)

    def compute_xr_peak_position(self, i, xr_params):
        return xr_params[i,1] + xr_params[i,3]



"""
    SecTheory.HermansEmg.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, root
from scipy.stats import linregress
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Peaks.ElutionModels import emg, egh
from LPM import get_corrected
from .FoleyDorseyEmg import fd_estimate_initial_params, convert_to_xr_params_fd

ESTIMATE_INJECTION_TIME = True
PENALTY_SCALE = 1e8
NEGTIVE_PENALTY_WEIGHT = 1000   # should be dynalically determined?

def compute_momemts_from_params(mu, sigma, tau):
    X = mu + tau
    Y = sigma**2 + tau**2
    Z = 2*tau**3
    return X, Y, Z

def compute_params_from_momemts(X, Y, Z):
    if np.isscalar(X):
        t = (Z/2)**(1/3)
        m = X - t
        s2 = Y - t**2
        s2_ = max(1, s2)
        s = np.sqrt(s2_)
        return m, s, t
    else:
        # s2_ = np.max([np.ones(len(s2)), s2], axis=0)
        params = np.zeros((len(X),3))
        for k, (x, y, z) in enumerate(zip(X, Y, Z)):
            params[k,:] = compute_params_from_momemts(x, y, z)
        return params

def compute_moments_from_ksec_p_r(t0, Ksec, P, R):
    K = Ksec*P
    X = t0 + K
    Y = 1/30 * K * R
    Z = 1/420 * K * R**2
    return X, Y, Z

def estimate_secparams_and_R(x, y, init_params, hs, trs, rgs, fdX, debug=False):
    """
    optimize secparams and R so that the model curve fits to the data
    """

    def objective(p, return_detail=False, debug=False, title=""):
        sec_params = p[0:-1]
        R = p[-1]
        X = convert_to_xr_params_hermans(hs, rgs, sec_params, R)
        ty = np.zeros(len(x))

        for i, (h, m, s, t) in enumerate(X):
            cy = emg(x, h, m, s, t)
            ty += cy

        if debug:
            ty_ = np.zeros(len(x))
            cy_list = []
            for i, (h, m, s, t) in enumerate(X):
                cy = emg(x, h, m, s, t)
                ty_ += cy
                if debug:
                    cy_list.append(cy)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y)
                for k, cy in enumerate(cy_list):
                    ax.plot(x, cy, ":", label="component-%d" % k)
                ax.plot(x, ty_, ":", color="red", label="total")
                ax.legend()
                fig.tight_layout()
                plt.show()

        fv1 = np.log10(np.sum((ty - y)**2))
        fv2 = np.log10(np.linalg.norm(X - fdX)**2)
        if return_detail:
            return fv1, fv2
        else:
            return  fv1 + fv2

    minP = trs[-1] - trs[0]
    R = init_params[-1]
    poresize_bounds = get_setting("poresize_bounds")
    bounds = [(-100, trs[0] - 10), (minP, minP*5), poresize_bounds, (1, 3), (R/2, R*2)]

    if debug:
        print("init_params=", init_params)
        print("bounds=", bounds)
        print("fv1, fv2=", objective(init_params, return_detail=True))
        objective(init_params, debug=True, title="estimate_secparams_and_R: init")

    ret = minimize(objective, init_params, bounds=bounds)

    if debug:
        print("ret.x=", ret.x)
        print("fv1, fv2=", objective(ret.x, return_detail=True))
        objective(ret.x, debug=True, title="estimate_secparams_and_R: optimized")

    return ret.x

def estimate_initial_params(ecurve, init_xr_params, init_rgs, seccol_params, D, debug=False):

    temp_xr_params, fd_seccol_params = fd_estimate_initial_params(ecurve, init_xr_params, seccol_params, debug=debug)
    Ti, Np = fd_seccol_params[-2:]
    fdX = convert_to_xr_params_fd(temp_xr_params, Ti, Np)

    x = ecurve.x
    y_ = ecurve.y

    # use corrected y
    y = get_corrected(y_)

    mu_ = np.average(init_xr_params[:,1])
    hs = init_xr_params[:,0]
    trs = init_xr_params[:,1]       # egh tr's

    t0, P, rp, m = seccol_params

    rhos =init_rgs/rp
    rhos[rhos > 1] = 1          # rhos[0] > 1 as in 20191114_4
    Ksec = (1 - rhos)**m

    slope, intercept = linregress(Ksec, trs)[0:2]

    t0 = intercept
    P = slope

    if debug:
        print("init_rgs=", init_rgs)
        print("rhos=", rhos)
        print("trs=", trs)
        print("t0, P=", t0, P)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("linregress")
            ax.plot(Ksec, trs, "o")
            ax.plot(Ksec, t0 + Ksec*P, "o")
            fig.tight_layout()
            plt.show()

    """
        M2 = 1/30*Ksec*P*R
        R = 30*M2/(Ksec*P)
    """
    sigma_ = np.average(init_xr_params[:,2])
    Ksec_ = np.average(Ksec)
    M2 = sigma_**2
    R = 30*M2/(Ksec_*P)

    init_params = [t0, P, rp, m, R]

    ret_params = estimate_secparams_and_R(x, y, init_params, hs, trs, init_rgs, fdX, debug=debug)
    secparams_ = ret_params[0:-1]
    R_ = ret_params[-1]

    if debug:
        t0_, P_, rp_, m = secparams_
        print("t0_, P_, rp_, m=", t0_, P_, rp_, m)
        rho_ = init_rgs/rp_
        ksec_ = (1 - rho_)**m
        tr = t0_ + ksec_*P_
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("SEC Parameters")
            ax.plot(x, y)
            for params  in init_xr_params:
                ax.plot(x, emg(x, *params), ":")
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)

            ax.plot([t0_, t0_], [ymin, ymax], ":", color="red")
            tp = t0_ + P_
            ax.plot([tp, tp], [ymin, ymax], ":", color="green")

            axt = ax.twinx()
            axt.grid(False)

            sy = np.linspace(0, rp_ - 0.1, 100)
            rho_ = sy/rp_
            sx = t0_ + P_*(1 - rho_)**m

            axt.plot(sx, sy, ":", color="blue")
            axt.plot(tr, init_rgs, "o", color="yellow")
            fig.tight_layout()
            plt.show()

    n = len(init_xr_params)
    p_zeros = np.zeros(D.shape[0])

    def objective(p, return_detail=False, debug=False, title=None):
        xr_params = convert_to_xr_params_hermans(p[0:n], init_rgs, p[n:-3], p[-1])
        ty = np.zeros(len(x))
        cy_list = []

        try:
            for h, m, s, t in xr_params:
                cy = emg(x, h, m, s, t)
                ty += cy
                cy_list.append(cy)

            C = np.array(cy_list)
            P = D @ np.linalg.pinv(C)
        except:
            """
            at ty += cy
            ValueError: operands could not be broadcast together with shapes (901,) (0,) (901,) 
            or numpy.linalg.LinAlgError: SVD did not converge
                where xr_params = 
                    [[nan nan  1. nan]
                     [nan nan  1. nan]
                     [nan nan  1. nan]]
            """
            if False:
                print("xr_params=", xr_params)
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("except")
                    ax.plot(x, y_, label="data")
                    ax.plot(x, y, label="corrected")
                    for k, cy in enumerate(cy_list):
                        ax.plot(x, cy, ":", label="component-%d" % k)
                    ax.plot(x, ty, ":", color="red", label="total")
                    fig.tight_layout()
                    plt.show()
            return PENALTY_SCALE

        negative_penalty = 0
        for p in P.T:
            negative_penalty += np.mean(np.min([p_zeros, p], axis=0)**2)

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                if title is None:
                    ax.set_title("objective")
                else:
                    ax.set_title(title)
                ax.plot(x, y)
                for cy in cy_list:
                    ax.plot(x, cy, ":")
                ax.plot(x, ty, ":", color="red")
                fig.tight_layout()
                plt.show()

        fv1 = np.log10(np.sum((ty - y)**2))
        fv2 = np.log10(np.linalg.norm(xr_params - fdX)**2)
        if return_detail:
            return fv1, fv2
        else:
            return fv1 + fv2 + NEGTIVE_PENALTY_WEIGHT*negative_penalty

    Np = get_setting("num_plates_pc")
    if ESTIMATE_INJECTION_TIME:
        """
            N = (m - Ti)**2/s**2
            sqrt(N) = (m - Ti)/s
            sqrt(N)*s = m - Ti
            Ti = m - sqrt(N)*s
        """
        i = np.argmax(init_xr_params[:,0])
        m_, s_ = init_xr_params[i,1:3]
        Ti = m_ - np.sqrt(Np)*s_
    else:
        from .ColumnConstants import INJECTION_TIME
        Ti = INJECTION_TIME

    init_values = np.concatenate([init_xr_params[:,0], secparams_, [Ti, Np, R_]])

    if debug:
        print("fv1, fv2=", objective(init_values, return_detail=True))
        objective(init_values, debug=True, title="init_values")

    bounds = [(0, None)]*len(init_values)
    ret = minimize(objective, init_values, bounds=bounds)
    temp_xr_params, seccol_params, R = ret.x[0:n], ret.x[n:n+4], ret.x[-1]

    if debug:
        print("ret.fun=", ret.fun)
        print("temp_xr_params=", temp_xr_params)
        print("seccol_params=", seccol_params)
        print("fv1, fv2=", objective(ret.x, return_detail=True))
        objective(ret.x, debug=True, title="ret.x")

    return temp_xr_params, seccol_params, R

def convert_to_xr_params_hermans(xrh_params, rgs, sec_params, R):
    t0, P, rp, m = sec_params
    rho = rgs/rp
    rho[rho > 1] = 1
    Ksec = (1 - rho)**m
    M = compute_moments_from_ksec_p_r(t0, Ksec, P, R)
    X = compute_params_from_momemts(*M)
    ret_params = np.zeros((len(xrh_params), 4))
    ret_params[:,0] = xrh_params
    ret_params[:,1:] = X
    return ret_params

def spike_demo():

    x = np.linspace(0, 500, 501)

    t0 = 50
    mu = 200
    sigma = 30
    tau = 10
    y = emg(x, 1, mu, sigma, tau)

    M = compute_momemts_from_params(mu, sigma, tau)
    m, s, t = compute_params_from_momemts(*M)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(x, emg(x, 1, m, s, t), ":")
        fig.tight_layout()
        plt.show()

    P = 300
    hs = np.array([1])
    trs = np.array([mu + tau])
    Ksec = 0.3
    M2 = sigma**2 + tau**2
    R = np.sqrt(30*M2/(Ksec*P))
    print("t0, P, R=", t0, P, R)

    t0_, P_, R_ = estimate_P_R(x, y, t0, P, R, hs, trs)
    print("t0_, P_, R=", t0_, P_, R_)
    K_ = (trs - t0_)/P_
    M = compute_moments_from_ksec_p_r(t0_, K_, P_, R_)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.plot(x, y)
        for i, (m_, s_, t_) in enumerate(compute_params_from_momemts(*M)):
            ax.plot(x, emg(x, hs[i], m_, s_, t_), ":")
        fig.tight_layout()
        plt.show()



"""

    UpdateDailog.py

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF

"""
from time import sleep
from molass_legacy.KekLib.KillableThread import Thread
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ScrolledText
from molass_legacy.KekLib.TkSupplements import set_icon, BlinkingFrame
from molass_legacy.KekLib.StdoutRedirector import StdoutRedirector

class UpdateDailog(Dialog):
    def __init__(self, parent, modules):
        self.modules = modules
        self.applied = False
        Dialog.__init__( self, parent, "PyTools Update", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):   # overrides Dialog.body
        set_icon(self)

        message_frame = Tk.Frame(body_frame)
        message_frame.pack(padx=20, pady=10)

        message =   (
                'The following modules are resuired\n'
                'to perform the function you have just requested.\n'
                'Press "Update" to install them into the embeddables\n'
                'before you proceed further.\n'
                'Cancelling here will result in cancelling the function.'
                )

        label = Tk.Label(message_frame, text=message, bg='white', anchor=Tk.W, justify=Tk.LEFT)
        label.pack()

        table_frame = Tk.Frame(body_frame)
        table_frame.pack(padx=20, pady=10)

        log_frame = Tk.Frame(body_frame)
        log_frame.pack(padx=20, pady=10)

        i = 0
        name_width = 10
        state_width = 40
        module_label = Tk.Label(table_frame, text="Name", width=name_width, relief=Tk.GROOVE)
        module_label.grid(row=i, column=0)
        module_state = Tk.Label(table_frame, text="State", width=state_width, relief=Tk.GROOVE)
        module_state.grid(row=i, column=1)

        self.state_frames = []
        for i, m in enumerate(self.modules, start=1):
            module_label = Tk.Label(table_frame, text=m, width=name_width, relief=Tk.GROOVE, bg="white")
            module_label.grid(row=i, column=0, sticky=Tk.W)
            state_frame = BlinkingFrame(table_frame)
            state_frame.grid(row=i, column=1, sticky=Tk.W)
            module_state = Tk.Label(state_frame, text="required to install", width=state_width,
                                    anchor=Tk.W, justify=Tk.LEFT, relief=Tk.GROOVE, bg="white")
            module_state.pack()
            state_frame.objects = [module_state]
            self.state_frames.append(state_frame)

        self.log_text = ScrolledText(log_frame, height=8)
        self.log_text.pack(fill=Tk.X)
        self.redirexctor = StdoutRedirector(self.log_text)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Update", width=10, command=self.update_modules, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.update_btn = w

        w = Tk.Button(box, text="OK", width=10, command=self.ok)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w.pack_forget()
        self.ok_btn = w

        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_btn = w

        self.bind("<Escape>", self.cancel)

    def apply(self):
        self.applied = True

    def update_modules(self):
        failed = False
        for k, m in enumerate(self.modules):
            state_frame = self.state_frames[k]
            state_frame.start()
            module_state = state_frame.objects[0]
            module_state.config(text="installing")
            self.start_install(m)
            self.wait_install_complete()
            state_frame.stop()
            if self.ret.returncode == 0:
                state = "installed"
                fg_color = 'black'
            else:
                state = "failed to install"
                failed = True
                fg_color = 'red'
            module_state.config(text=state, fg=fg_color)

        if not failed:
            self.update_btn.pack_forget()
            self.cancel_btn.pack_forget()
            self.ok_btn.pack(side=Tk.LEFT, padx=5, pady=5)

    def start_install(self, m):
        self.install_thread = Thread(
                        target=self.install,
                        name='InstallThread',
                        args=[m]
                        )
        self.install_complete = False
        self.install_thread.start()

    def install(self, m):
        import sys
        import subprocess
        from SubProcess import get_startup_info
        print('installing', m)
        python  = sys.executable.replace('pythonw.exe', 'python.exe')
        ret = subprocess.run([python, '-m', 'pip', 'install', '-U', m], capture_output=True, startupinfo=get_startup_info())
        stdout = ret.stdout.decode()
        stderr = ret.stderr.decode()
        print(stdout)
        if ret.returncode != 0:
            print(stderr)
        self.install_complete = True
        self.ret = ret

    def wait_install_complete(self):
        while not self.install_complete:
            self.update()
            sleep(0.1)



"""
module for the usual method of moments
"""
import sympy
import numpy as np
import scipy
from discrete_rv import DiscreteRV
from model_gm import ModelGM
import moments as mom

class MM():
    """
    class for the usual method of moments
    sigma = 1
    """
    def __init__(self, k, sigma=1):
        self.k = k
        self.sigma = sigma
        self.num_mm = 2*k-1
        self.transform = mom.hermite_transform_matrix(2*self.k-1, self.sigma)

    def estimate(self, samples):
        """
        estimate from the usual method of moments
        """
        samples = np.asarray(samples)
        m_raw = mom.empirical_moment(samples, 2*self.k-1)
        m_raw = np.mean(m_raw, axis=1).reshape((self.num_mm, 1))
        m_decon = np.dot(self.transform[0], m_raw)+self.transform[1]
        mm_rv = self.mom_symbol(m_decon.reshape(self.num_mm))
        return ModelGM(w=mm_rv.weights, x=mm_rv.atoms, std=self.sigma)

    def mom_symbol(self, moments):
        """ Symbolic solver for ordinary method of moments
        find k-point representative distribution

        Args:
        moments: a sequence of estimated moments, starting from degree 1
            only use first 2k-1 moments
        num_comps: number of atoms in the output RV

        Returns:
        U(finiteRV): estimated RV, whose first 2k-1 moments equal m

        """

        num = 2*self.k-1
        assert len(moments) >= num

        p_var = sympy.symbols('p0:%d'%(self.k), nonnegative=True, real=True)
        x_var = sympy.symbols('x0:%d'%(self.k), real=True)

        eqn = -1
        for i in range(self.k):
            eqn = eqn + p_var[i]
        equations = [eqn]

        for i in range(num):
            eqn = -moments[i]
            ########### truncate a real number to rational
            # eq = -nsimplify(m[i], tolerance=0.1)
            # eq = float('%.2g' % -m[i])
            for j in range(self.k):
                eqn = eqn + p_var[j]*(x_var[j]**(i+1))
            equations.append(eqn)

        var = [x for xs in zip(p_var, x_var) for x in xs] # format: [p1,x1,p2,x2,...]
        s_var = sympy.solve(equations, var)
        # s = solveset(equations,list(p)+list(x))
        # s = nonlinsolve(equations,list(p)+list(x))

        ########## some other tools in SymPy
        # print(equations)
        # s = solve_poly_system(equations,list(p)+list(x))
        # s = solve_triangulated(equations,list(p)+list(x))
        # s = solve_triangulated(equations,p[0],p[1],x[0],x[1])

        ########## test over simple rational coefficients
        # testEq = [x[0]*p[0] - 2*p[0], 2*p[0]**2 - x[0]**2]
        # s = solve_triangulated(testEq, x[0], p[0])

        if len(s_var) == 0:
            return DiscreteRV(w=[], x=[])
        else:
            return DiscreteRV(w=s_var[0][0::2], x=s_var[0][1::2])



    def mom_numeric(self, moments, start):
        """ numerical solver for ordinary method of moments
        find k-point representative distribution

        Args:
        moments: a sequence of estimated moments, starting from degree 1
            only use first 2k-1 moments
        start: initial guess. k = the number of components in start.

        Returns:
        U(finiteRV): estimated RV

        """
        k = len(start.p)
        assert k == self.k

        num = 2*k-1
        assert len(moments) >= num
        moments = moments[0:num]

        def equation(x_var):
            """
            local moments condition equation
            """
            model = DiscreteRV(x=x_var[0:k], w=np.append(x_var[k:], 1-np.sum(x_var[k:])))
            return model.moment(num)-moments

        x0_var = np.concatenate((start.x, start.p[0:-1]))
        x_var = scipy.optimize.fsolve(equation, x0_var)
        # x = scipy.optimize.broyden2(equation, x0)

        return DiscreteRV(x=x_var[0:k], w=np.append(x_var[k:], 1-np.sum(x_var[k:])))



# coding: utf-8
"""
    DataSubtractor.py

    Copyright (c) 2019-2021, SAXS Team, KEK-PF
"""
import os
import numpy as np
import difflib
from SerialDataUtils import load_xray_files
from molass_legacy.KekLib.NumpyUtils import np_savetxt_with_comments
from molass_legacy._MOLASS.SerialSettings import set_setting, get_setting
import molass_legacy.KekLib.CustomMessageBox as MessageBox

def subtract_impl(in_files1, in_files2, process_option, out_folder, name_pattern, dialog=None):

    set_setting('found_lacking_q_values', 0)
    data_array1, comments1, lacking_info1 = load_xray_files(in_files1, return_lacking_info=True)
    if get_setting('found_lacking_q_values'):
        MessageBox.showwarning( "Lacking Q-values",
            "There have been found lacking Q-values\n"
            "in some of the input files.\n"
            "Make sure to confirm them in the molass.log",
            parent=dialog )

    if process_option == 0:
        minn, maxn, lacking_list, minn_qv = lacking_info1
        qv = minn_qv if len(lacking_list) > 0 else None
        data_array2, comments2, lacking_info2 = load_xray_files(in_files2, return_lacking_info=True, qv=qv)
        last_line = comments1.pop(-1)
        comments2.pop(-1)

        comments1.append("### comments difference from the subtracted begin ###\n")
        for line in difflib.unified_diff(comments1, comments2):
            if line[0:2] == '+#':
                # print(line)
                comments1.append(line[1:])
        comments1.append("### comments difference from the subtracted end ###\n")
        comments1.append(last_line)
    else:
        average2 = in_files2

    if dialog is not None:
        dialog.mpb["maximum"] = len(in_files1)
        dialog.mpb["value"] = 0

    out_data_list = []
    for k in range(len(in_files1)):
        data1 = data_array1[k,:,:]
        if process_option == 0:
            data2 = data_array2[k,:,:]
        else:
            data2 = average2
        out_data = np.vstack([data1[:,0], data1[:,1] - data2[:,1], np.sqrt(data1[:,2]**2 + data2[:,2]**2)]).T
        filename = name_pattern % k
        np_savetxt_with_comments(os.path.join(out_folder, filename), out_data, comments1)
        out_data_list.append(out_data)

        if dialog is not None:
            dialog.num_files3.set(k+1)
            dialog.mpb["value"] = k+1
            dialog.update()

    return out_data_list



"""

    AutoRunTestTools.py

    Copyright (c) 2020, SAXS Team, KEK-PF

"""
import os
import glob
from time import sleep
from datetime import datetime, timedelta, timezone
import numpy as np
from bisect import bisect_left
from shutil import copy

JST = timezone(timedelta(hours=+9), 'JST')

def my_strftime(t):
    return t.strftime('%Y-%m-%d %H:%M:%S')

class RscSimulator:
    def __init__(self, in_folder, out_folder):
        self.in_folder = in_folder
        self.out_folder = out_folder
        assert os.path.exists(self.out_folder) and os.path.isdir(self.out_folder)
        self.get_in_folder_info()

    def get_in_folder_info(self):
        measurement_log = None
        for k, path in enumerate(glob.glob(self.in_folder + r'\*.log')):
            if path.find('measurement') > 0:
                print([k], path)
                measurement_log = path

        mask_file = None
        for k, path in enumerate(glob.glob(self.in_folder + r'\*.mask')):
            print([k], path)
            mask_file = path

        pilatus_counters = []
        for k, path in enumerate(glob.glob(self.in_folder + r'\*.txt')):
            if path.find('PilatusCounter') > 0:
                print([k], path)
                pilatus_counters.append(path)

        self.measurement_log = measurement_log
        self.mask_file = mask_file
        self.pilatus_counters = pilatus_counters

        time_list = []
        file_list = []
        for ext in ['tif', 'cbf']:
            for k, path in enumerate(glob.glob(self.in_folder + r'\*.' + ext)):
                # print([k], path)
                stat_result = os.stat(path)
                if False:
                    for t in [stat_result.st_atime, stat_result.st_mtime, stat_result.st_ctime]:
                        print(datetime.fromtimestamp(t, JST).strftime('%Y-%m-%d %H:%M:%S.%f'))
                m_time = datetime.fromtimestamp(stat_result.st_mtime, JST)
                time_list.append(m_time)
                file_list.append(path)
                print([k], my_strftime(m_time), path)

        print(len(file_list))
        self.time_list = np.array(sorted(time_list))
        self.file_list = file_list

    def run(self, interval=10, timescale=6):
        for path in [self.measurement_log, self.mask_file] + self.pilatus_counters:
            if path is None:
                print("ERROR: One of required path is None!")
                print("Make sure the existence of measurement.log, mask_file, or pilatus_counters.")                
                assert False, "See the measage above."
            print(path)
            copy(path, self.out_folder)

        self.start_time = self.time_list[0] - timedelta(seconds=1)
        print(my_strftime(self.time_list[0]))
        print(my_strftime(self.start_time))

        copy_start = 0
        for step in range(1000):
            sleep(interval)
            current_time = self.start_time + timedelta(seconds=step*interval*timescale)
            print('step %d' % step, my_strftime(current_time))
            k = bisect_left(self.time_list, current_time)
            for j in range(copy_start, k):
                print('copy', [j], my_strftime(self.time_list[j]))
                path  = self.file_list[j]
                copy(path, self.out_folder)
            copy_start = k



# -*- coding: utf-8 -*-
"""

    ファイル名：   TestTkUtils.py

    処理内容：

        

"""
from __future__ import division, print_function, unicode_literals

import os
import sys
import re

def split_geometry( geometry ):
    n = geometry.split('+')
    s = n[0].split('x')
    w = int( s[0] )
    h = int( s[1] )
    x = int( n[1] )
    y = int( n[2] )
    return [ w, h, x, y ]

def join_geometry( w, h, x, y ):
    return '%dx%d+%d+%d' % ( w, h, x, y )

def geometry_fix( top, x, y ):
    W, H, X, Y = split_geometry( top.geometry() )
    top.geometry( join_geometry( W, H, x, y ) )
    top.update()

"""
    URL: http://effbot.org/tkinterbook/wm.htm
"""
def parsegeometry(geometry):
    m = re.match("(\d+)x(\d+)([-+]\d+)([-+]\d+)", geometry)
    if not m:
        raise ValueError("failed to parse geometry string")
    return map(int, m.groups())



# coding: utf-8
"""

    Hyperbola.py

    Copyright (c) 2021, SAXS Team, KEK-PF

"""
import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt

"""
    see rotated-hyperbola.ipynb
"""

def compute_rotated_hyperbola_y(x, ptx, pty, y0, a, b, theta):
    x_ = x - ptx
    y_ = -a/b * np.sqrt(x_**2 + b**2) + y0 - pty
    return x_*np.sin(theta) + y_*np.cos(theta) + pty        

HBW_SCALE = 30

class RotatedHyperbola:
    def __init__(self, ptx, pty, a=None, b=None, hbw=None,deg=0):
        assert (b is not None) or (hbw is not None)
        if a is None:
            a = pty/hbw*HBW_SCALE
        y0 = pty + a
        if b is None:
            b = hbw/np.sqrt((y0/a)**2 - 1)
        self.ptx = ptx
        self.pty = pty
        self.x0 = ptx
        self.y0 = y0
        self.a = a
        self.b = b
        theta = deg*np.pi/180
        self.init_params = (ptx, pty, y0, a, b, theta)
        self.fitted_params = None
        self.spline = None

    def fit(self, x, y, method=None):
        def obj_func(p):
            return np.sum((compute_rotated_hyperbola_y(x, *p) - y)**2)
        result = minimize(obj_func, self.init_params, method=method)
        self.fitted_params = result.x
        print("self.fitted_params=", self.fitted_params)

    def __call__(self, x):
        params = self.init_params if self.fitted_params is None else self.fitted_params
        return compute_rotated_hyperbola_y(x, *params)

    def get_peak_top(self):
        ptx, pty, _, a = self.fitted_params[0:4]
        hbw = pty/a*HBW_SCALE
        x_ = np.linspace(ptx-hbw, ptx+hbw, 100)
        y_ = compute_rotated_hyperbola_y(x_, *self.fitted_params)
        n = np.argmax(y_)
        return np.array([x_[n], y_[n]])



"""
    Theory.StructureFactor.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.integrate import quad
import molass_legacy.KekLib.DebugPlot as plt


def demo1():

    Na_pos = np.array([(0,0,0), (0,1/2,1/2), (1/2,0,1/2), (1/2,1/2,0)])
    Cl_pos = np.array([(1/2,0,0), (0,1/2,0), (0,0,1/2), (1/2,1/2,1/2)])

    fig = plt.figure(figsize=(14,7))

    ax1 = fig.add_subplot(121, projection='3d')
    ax2 = fig.add_subplot(122)

    ax1.scatter(*Na_pos.T, color='blue')
    ax1.scatter(*Cl_pos.T, color='red')

    fig.tight_layout()
    plt.show()



"""
    TheDebugUtils.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Elution.CurveUtils import simple_plot
from molass_legacy.Optimizer.BasicOptimizer import SCORE_DEV_ALLOW, SCORE_DEV_LIMIT
from molass_legacy._MOLASS.SerialSettings import get_setting

def plot_mapped_curve(ax, map_params, x, uv_curve):
    A, B = map_params
    uv_y = uv_curve.spline(A*x + B)
    ax.plot(x, uv_y, color='blue', label='data')
    topx_list = []
    for k, info in enumerate(uv_curve.get_major_peak_info()):
        topx = info[1]
        topx_list.append(topx)
        topy = uv_curve.spline(topx)
        x_ = (topx - B)/A
        label = 'peak tops' if k == 0 else None
        ax.plot(x_, topy, 'o', color='red', label=label)

    for k, btmx in enumerate(uv_curve.get_major_valley_bottoms(topx_list)):
        btmy = uv_curve.spline(btmx)
        x_ = (btmx - B)/A
        label = 'valley bottoms' if k == 0 else None
        ax.plot(x_, btmy, 'o', color='green', label=label)

def plot_decomp_elements(ax, x, opt_recs):
    for k, rec in enumerate(opt_recs):
        fnc = rec[1]
        cy = fnc(x)
        ax.plot(x, cy, ':', label='component-%d' % k)

def plot_decomp_result_titles(fig, ax1, ax2):
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
    in_folder = get_in_folder()
    fig.suptitle("Initial Parameters Estimation for %s" % in_folder, fontsize=20)
    ax1.set_title("Mapped UV Decomposition", fontsize=16)
    ax2.set_title("Xray Decomposition", fontsize=16)

def plot_decomp_result(uv_curve, xr_curve, map_params, decomp_result=None, fig_info=None):

    if fig_info is None:
        plt.push()
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
    else:
        fig, (ax1, ax2) = fig_info

    plot_decomp_result_titles(fig, ax1, ax2)

    x = xr_curve.x
    plot_mapped_curve(ax1, map_params, x, uv_curve)
    simple_plot(ax2, xr_curve, color='orange', legend=False)

    if decomp_result is not None:
        plot_decomp_elements(ax1, x, decomp_result.opt_recs_uv)
        plot_decomp_elements(ax2, x, decomp_result.opt_recs)

    for ax in (ax1, ax2):
        ax.legend()

    if fig_info is None:
        fig.tight_layout()
        fig.subplots_adjust(top=0.9)
        plt.show()
        plt.pop()

def convert_score_list(score_list_pair):
    score_list, penalties = score_list_pair
    return np.concatenate([score_list, penalties])

def plot_scores(ax, score_list, score_names, alpha=1, label=None, invert=True, add_patch=True, debug=True):
    if debug:
        from importlib import reload
        import molass_legacy.Optimizer.FvScoreConverter
        reload(molass_legacy.Optimizer.FvScoreConverter)
    from .FvScoreConverter import convert_score

    converted_list = [convert_score(v) for v in score_list]

    NUM_MAJOR_SCORES = get_setting("NUM_MAJOR_SCORES")
    major_scores = score_list[0:NUM_MAJOR_SCORES]
    score_mean = np.mean(major_scores)
    score_dev = max(SCORE_DEV_LIMIT, SCORE_DEV_ALLOW * np.std(major_scores))
    f = convert_score(score_mean - score_dev)
    t = convert_score(score_mean + score_dev)

    ax.yaxis.tick_right()
    ax.yaxis.set_label_position("right")
    y_pos = np.arange(len(score_list))
    ax.barh(y_pos, converted_list, label=label, alpha=alpha)
    ax.set_yticks(y_pos, score_names)

    if invert:
        ax.invert_yaxis()

    xmin, xmax = ax.get_xlim()
    ax.set_xlim(xmin, 105)

    if add_patch:
        ymin, ymax = ax.get_ylim()
        p = Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.1,
            )
        ax.add_patch(p)



"""

    FileTimeUtils.py

    Copyright (c) 2019, Masatsuyo Takahashi, KEK-PF

    References

    How do I change the file creation date of a Windows file from Python?
    https://stackoverflow.com/questions/4996405/how-do-i-change-the-file-creation-date-of-a-windows-file-from-python

    
    
"""
import os
import win32file, win32con, pywintypes

def getFileDateTimes( filePath ):        
    return ( os.path.getctime( filePath ), 
             os.path.getmtime( filePath ), 
             os.path.getatime( filePath ) )

def setFileDateTimes( filePath, datetimes ):
    ctime, mtime, atime = [pywintypes.Time(t) for t in datetimes]
    winfile = win32file.CreateFile(
        filePath, win32con.GENERIC_WRITE,
        win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE | win32con.FILE_SHARE_DELETE,
        None, win32con.OPEN_EXISTING,
        win32con.FILE_ATTRIBUTE_NORMAL, None)
    win32file.SetFileTime( winfile, ctime, atime, mtime )
    winfile.close()



# coding: utf-8
"""
    SimpleGuinierScore.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy        as np
from scipy          import stats

MIN_GUINIER_SIZE        =  8            # required for Matsumura2-173 with QUALITY_BOUNDARY_QRG==0.5
SUF_GUINIER_SIZE        = 60            # required for Matsumura2-179, 180, 211 when END_GUINIER_SIZE==20
FWD_GUINIER_SIZE        = 10
END_GUINIER_SIZE        = 20
LINEARITY_SCALE         = 5
RG_CONSISTENCY_SCALE    = 20
QUALITY_BOUNDARY_IWD    = 0.1
IWD_CONSISTENCY_LIMIT   = 0.3
QUALITY_BOUNDARY_LOW    = 0.2
QUALITY_BOUNDARY_MID    = 0.7
QUALITY_BOUNDARY_HIGH   = 0.9
CONSISTENCY_BOUNDARY    = 0.5
USE_FWD_CONSISTENCY     = True

if USE_FWD_CONSISTENCY:
    SCORING_WEIGHTS     = np.array( [ 0.4, 0.5, 0.06, 0.04 ] )
    SCORING_WEIGHTS_MH  = np.array( [ 0.6, 0.3, 0.06, 0.04 ] )
    SCORING_WEIGHTS_ML  = np.array( [ 0.7, 0.3, 0.0, 0.0 ] )        # Matsumura2-198
    SCORING_WEIGHTS_LQ  = np.array( [ 0.8, 0.2, 0.0, 0.0 ] )
else:
    SCORING_WEIGHTS = np.array( [ 0.5, 0.4, 0.1 ] )
NUM_SCORE_FACTORS       = len(SCORING_WEIGHTS)
ZERO_SCORE_VECTOR = np.zeros(NUM_SCORE_FACTORS)
LOGARITHMIC_SCORING     = True
USE_GUINIER_PLOT_LENGTH = False         # should be True for 20160227-extrapolated-asc-003

DEBUG_PLOT = False

def compute_rg( slope ):
    if slope < 0:
        rg = np.sqrt( -3*slope )
    else:
        rg = 0
    return rg

def evaluate_rg_consistency( rg1, rg2 ):
    return 0 if rg2 == 0 else np.exp( -abs( 1 - rg1/rg2 )*RG_CONSISTENCY_SCALE )

def compute_fwd_consistency( start, stop, x2, log_y ):
    if not USE_FWD_CONSISTENCY:
        return None

    from_ = max( 0, start - FWD_GUINIER_SIZE )
    to_ = max( from_ + FWD_GUINIER_SIZE, start )

    rgs = []
    for slice_ in [ slice( from_, to_), slice(stop, stop + FWD_GUINIER_SIZE) ]:
        x_ = x2[slice_]
        y_ = log_y[slice_]
        slope = stats.linregress( x_, y_ )[0]
        rgs.append( compute_rg( slope ) )
    # print( 'compute_end_consistency: rgs=', rgs )
    return evaluate_rg_consistency( rgs[0], rgs[1] )

def compute_end_consistency( start, stop, x2, log_y, debug=False ):
    end_size = min( (stop-start)//2, END_GUINIER_SIZE )
    slice_ = slice( stop - end_size, stop )
    rgs = []

    debug_ = DEBUG_PLOT or debug

    if debug_:
        debug_intervals = []

    for slice_ in [ slice(start, start+end_size), slice(stop-end_size, stop) ]:
        x_ = x2[slice_]
        y_ = log_y[slice_]
        slope = stats.linregress( x_, y_ )[0]
        rgs.append( compute_rg( slope ) )
        if debug_:
            debug_intervals.append( [ x_, y_ ] )
    # print( 'compute_end_consistency: rgs=', rgs )

    if debug_:
        from CanvasDialog   import CanvasDialog
        slice_ = slice( 0, stop )
        x   = x2[slice_]
        y   = log_y[slice_]
        print( 'rgs=', rgs )
        def func( fig ):
            ax = fig.add_subplot(111)
            ax.set_title( 'compute_end_consistency')
            ax.plot( x, y, 'o', markersize=3 )
            for x_, y_ in debug_intervals:
                ax.plot( x_, y_, 'o', markersize=3 )
            fig.tight_layout()
        dialog = CanvasDialog( "Debug" )
        dialog.show( func, figsize=(9,8), toolbar=True )

    return evaluate_rg_consistency( rgs[0], rgs[1] )

def get_linearity( score_array ):
    # return np.prod( [ np.power( score_array[k+1], w ) for k, w in enumerate( [0.5, 0.5]) ] )
    return np.sum( [ score_array[k+1] * w for k, w in enumerate( [0.5, 0.5]) ] )

def compute_expected_size( px, dx, rg ):
    max_q = 1.3/rg - px
    expected_size = max( MIN_GUINIER_SIZE+1, int( max_q/dx ) )
    # max_size = min( SUF_GUINIER_SIZE, expected_size )
    return expected_size

def compute_expected_length( px, rg ):
    max_q = 1.3/rg
    dx  = max_q**2 - px**2      # width  in Guinier plot scale
    dy  = dx * (rg**2) /3       # height in Guinier plot scale
    length = np.sqrt( dx**2 + dy**2 )
    return length

def evaluate_guinier_interval( bq, peak, dx, rg, length, r_value, end_consistency, fwd_consistency, spline_rg, return_vector=False, debug=False ):
    if USE_GUINIER_PLOT_LENGTH:
        max_length  = compute_expected_length( peak, rg )
    else:
        max_length  = compute_expected_size( peak, dx, rg )
    if debug:
        print( 'peak=', peak, 'dx=', dx, 'rg=', rg, 'length=', length, 'max_length=', max_length )

    if  bq < QUALITY_BOUNDARY_LOW:
        weights         = SCORING_WEIGHTS_LQ
    elif bq > QUALITY_BOUNDARY_HIGH:
            weights     = SCORING_WEIGHTS
    else:
        if bq > QUALITY_BOUNDARY_MID:
            # print( 'end_consistency=', end_consistency )
            if end_consistency < CONSISTENCY_BOUNDARY:
                # neccesary e.g. for 20160227
                weights = SCORING_WEIGHTS
            else:
                weights = SCORING_WEIGHTS_MH
        else:
            weights = SCORING_WEIGHTS_ML

    if spline_rg is None:
        ratio = min( 1, length/max_length )
        size_score = ratio**2
    else:
        size_score = evaluate_rg_consistency( rg, spline_rg )

    if debug:
        print( 'size_score=', size_score )

    linearity_score = np.exp( -(r_value + 1)*LINEARITY_SCALE )

    if USE_FWD_CONSISTENCY:
        score_array = np.array( [size_score, linearity_score, end_consistency, fwd_consistency ] )
    else:
        score_array = np.array( [size_score, linearity_score, end_consistency ] )

    score_vector = weights * score_array

    if LOGARITHMIC_SCORING:
        score = np.prod( [ np.power( score_array[k], w ) for k, w in enumerate( weights ) ] )
    else:
        score = np.sum( score_vector )
    if return_vector:
        return score, score_vector, score_array
    else:
        return score



"""
    Theory.EllisoidPlot.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.integrate import quad
from scipy.special import spherical_jn
from .RotationMatrix import rand_rotation_matrix, rotate_3d_mesh
import molass_legacy.KekLib.DebugPlot as plt


def ellipsoid_xyz(center=(0,0,0), a=1, b=1, c=1, n=20):
    """
    https://matplotlib.org/mpl_toolkits/mplot3d/tutorial.html
    """
    u = np.linspace(0, 2 * np.pi, n)
    v = np.linspace(0, np.pi, n)
    x = a * np.outer(np.cos(u), np.sin(v))
    y = b * np.outer(np.sin(u), np.sin(v))
    z = c * np.outer(np.ones(np.size(u)), np.cos(v))
    return center[0]+x, center[1]+y, center[2]+z

def demo1(seed=1111):

    fig = plt.figure(figsize=(10,10))
    ax = fig.add_subplot(111, projection='3d')

    R = 0.3
    Rg = np.sqrt(3/5)*R
    # radii = np.random.uniform(0.5, 1.5, 3)*R
    radii = np.array([2, 3, 4])
    scale = Rg/np.sqrt(np.sum(radii**2)/5)
    radii_ = radii*scale

    v0 = 4*np.pi*np.prod(radii_)/3
    ax.set_title(r"$N=16, V=4^3=64, v_1=\frac{V}{N}=4, v_0=\frac{4}{3}\pi a b c  =%.3g$" % v0, fontsize=20)

    np.random.seed(seed)
    purtubations = np.random.uniform(-0.5, 0.5, (16,3))
    for k, center in enumerate([
                    (0.5, 0.5, 0.5), (2.5, 0.5, 0.5), (0.5, 0.5, 2.5), (2.5, 0.5, 2.5),
                    (1.5, 1.5, 1.5), (3.5, 1.5, 1.5), (1.5, 1.5, 3.5), (3.5, 1.5, 3.5), 
                    (0.5, 2.5, 0.5), (2.5, 2.5, 0.5), (0.5, 2.5, 2.5), (2.5, 2.5, 2.5),
                    (1.5, 3.5, 1.5), (3.5, 3.5, 1.5), (1.5, 3.5, 3.5), (3.5, 3.5, 3.5), 
                    ]):
        p = purtubations[k]
        center_ = center+p
        s = ellipsoid_xyz(center_, *radii_)
        M = rand_rotation_matrix()
        rs = rotate_3d_mesh(M, center_, *s)
        ax.plot_surface(*rs, color='b')

    x, y, z = np.indices((4, 4, 4))
    cubes = x + y + z  >= 0
    colors = np.array([[['#1f77b430']*4]*4]*4)

    ax.voxels(cubes, facecolors=colors)

    fig.tight_layout()
    plt.show()



# coding: utf-8
"""
    Scattering3dSimple.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import copy
import numpy                as np
from bisect                 import bisect_right
from mpl_toolkits.mplot3d   import Axes3D, proj3d
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.TkUtils                import is_low_resolution, split_geometry
from CanvasFrame            import CanvasFrame
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting, get_xray_picking

VERY_SMALL_ANGLE_LIMIT  = 0.03
NUM_SAMPLE_CURVES       = 20
ZLIM_EXPAND             = 3.0

def is_to_plot( index, i, q, very_small_angle_only=False ):
    # return q >= 0.01 and ( i == self.index or i % 8 == 0 )
    if very_small_angle_only:
        return q < VERY_SMALL_ANGLE_LIMIT or i == index
    else:
        return i == index or i % 8 == 0

class Scattering3dSimpleFrame( Tk.Frame ):
    def __init__( self, parent, dialog, serial_data, data, title, i ):
        experimental = i == 1
        self.plot_no        = i
        self.parent         = parent
        self.dialog         = dialog
        self.qvector        = serial_data.qvector
        self.jvector        = serial_data.jvector
        self.index          = bisect_right( self.qvector, get_xray_picking() )
        self.xray_curve_y   = serial_data.xray_curve.y
        self.xray_slice     = serial_data.xray_slice
        self.data           = data
        self.data_list      = [ data ]
        self.title          = title
        self.el_x   = np.ones( len(self.jvector) ) * self.qvector[self.index]
        self.el_y   = self.jvector
        self.el_z   = self.xray_curve_y
        self.suppress_on_draw   = False

        Tk.Frame.__init__( self, parent )

        self.body( self )
        self.buttonbox()
        self.draw()
        self.create_popup_menu()

    def compute_baselines( self, corrector ):

        vsa_base_list = []
        all_base_list = []
        for i, q  in enumerate( self.qvector ):
            i_select = ( i == self.index or i % 8 == 0 )
            if q < VERY_SMALL_ANGLE_LIMIT or i_select:
                baseline = corrector.correct_a_single_q_plane( i, return_baseline=True )
                if q < VERY_SMALL_ANGLE_LIMIT or i == self.index:
                    vsa_base_list.append( baseline )
                if i_select:
                    all_base_list.append( baseline )

        return vsa_base_list, all_base_list

    def body( self, body_frame ):   # overrides parent class method

        base_frame = Tk.Frame( body_frame );
        # base_frame.pack( expand=1, fill=Tk.BOTH, padx=20, pady=10 )
        base_frame.pack( expand=1, fill=Tk.BOTH )

        figsize = ( 8, 8 ) if is_low_resolution() else ( 10, 10 )

        self.canvas_frame = canvas_frame = CanvasFrame( base_frame, figsize=figsize )
        canvas_frame.pack()
        fig = self.canvas_frame.fig
        ax1 = fig.add_subplot( 111, projection='3d' )
        self.axes = [ ax1 ]

        fig.tight_layout()
        self.canvas_frame.fig.canvas.mpl_connect( 'draw_event', self.on_draw )
        self.canvas_frame.fig.canvas.mpl_connect( 'button_press_event', self.on_button_press )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        frame = self

        box1 = Tk.Frame( frame )
        box1.pack( fill=Tk.X, anchor=Tk.CENTER )
        self.button_frame = Tk.Frame( box1 )
        self.button_frame.grid( row=0, column=0 )
        box1.columnconfigure( 0, weight=1 )
        # box1.columnconfigure( 1, weight=1 )

        self.cunstom_button_box_add( self.button_frame )

    def cunstom_button_box_add( self, frame ):
        self.vsa_only = Tk.IntVar()
        self.vsa_only.set( 1 )
        self.vsa_only_cb = Tk.Checkbutton( frame, text="Very Small Angle only", variable=self.vsa_only )
        self.vsa_only_cb.grid( row=0, column=0, padx=5 )

        space = Tk.Frame( frame, width=10 )
        space.grid( row=0, column=1, padx=10 )

        label = Tk.Label( frame, text="Zlim expand ratio:" )
        label.grid( row=0, column=2 )
        self.zlim_expand = Tk.DoubleVar()
        self.zlim_expand.set( ZLIM_EXPAND )
        entry = Tk.Entry( frame, textvariable=self.zlim_expand, width=6, justify=Tk.RIGHT )
        entry.grid( row=0, column=3 )

        space = Tk.Frame( frame, width=10 )
        space.grid( row=0, column=6, padx=10 )

        self.redraw_btn = Tk.Button( frame, text="Redraw", command=self.redraw )
        self.redraw_btn.grid( row=0, column=4, padx=10 )

        view_init = Tk.Button( frame, text="View Init", command=self.view_init )
        view_init.grid( row=0, column=5, padx=10 )

    def create_popup_menu( self ):
        self.popup_menu = Tk.Menu( self, tearoff=0 )
        self.popup_menu.add_command( label='Guinier Analysis', command=lambda: self.guinier_analysis() )
        self.popup_menu.add_command( label='Guinier Analysis Corrected', command=lambda: self.guinier_analysis(corrected=True) )

        # self.draw()
    def draw( self ):
        try:
            zlim_expand = self.zlim_expand.get()
        except:
            zlim_expand = None

        self.draw_3d_scattering( 0, self.title, zlim_expand=zlim_expand )
        self.canvas_frame.show()

    def redraw( self ):
        for ax in self.axes:
            ax.cla()
        self.draw()
        self.canvas_frame.show()

    def view_init( self ):
        for ax in self.axes:
            ax.view_init()
        self.canvas_frame.show()

    def draw_3d_scattering( self, n, title, zlim=None, zlim_expand=None ):
        ax      = self.axes[n]
        data    = self.data_list[n]

        very_small_angle_only = self.vsa_only.get() == 1

        ax.set_title( title )
        ax.set_xlabel( '\nQ(Å⁻¹)' )
        ax.set_ylabel( '\nElution №' )
        ax.set_zlabel( '\nIntensity' )
        # ax.zaxis._set_scale('log')

        if zlim is None:
            if zlim_expand is None:
                zlim_expand = ZLIM_EXPAND
            zmin    = np.min( self.xray_curve_y )
            zmax    = np.max( self.xray_curve_y )
            # zmin, zmax = ax.get_zlim()
            zmin_   = zlim_expand * zmin + ( 1-zlim_expand ) * zmax
            zmax_   = ( 1-zlim_expand ) * zmin + zlim_expand * zmax
            zlim    = ( zmin_, zmax_ )

        ax.set_zlim( zlim )

        print( 'data.shape=', data.shape )

        size = data.shape[0]
        Y = np.arange( size )

        peak_j = None
        i_ = -1
        i_list = []
        for i, q in enumerate( self.qvector ):
            if not is_to_plot( self.index, i, q, very_small_angle_only=very_small_angle_only ):
                continue

            i_list.append( i )

            X = np.ones( size ) * q

            if i == self.index:
                alpha   = 1
                color   = 'orange'
                Z = self.xray_curve_y
            else:
                alpha   = 0.2
                color   = '#1f77b4'
                Z = data[:,i,1]

            ax.plot( X, Y, Z, color=color, alpha=alpha )

            i_ += 1

            if False:
                if very_small_angle_only:
                    baseline = vsa_base_list[i_]
                else:
                    baseline = all_base_list[i_]
                ax.plot( X, Y, baseline, color='red', alpha=0.2 )

        return

    def on_draw( self, event ):
        # print( 'on_draw' )
        if self.dialog.sync_view.get() == 0:
            return

        if self.suppress_on_draw:
            return

        this_ax = self.axes[0]
        other_frame = self.dialog.plot_frames[1-self.plot_no]
        # https://stackoverflow.com/questions/47610614/get-viewing-camera-angles-in-matplotlib-3d-plot
        other_frame.suppress_on_draw = True

        zlim_expand_ = self.zlim_expand.get()
        if other_frame.zlim_expand.get() != zlim_expand_:
            other_frame.zlim_expand.set(zlim_expand_)
            other_frame.redraw()

        other_frame.axes[0].view_init( this_ax.elev, this_ax.azim )
        other_frame.canvas_frame.show()
        other_frame.update()

        other_frame.suppress_on_draw = False

    def compute_projected_xy( self ):
        ax = self.axes[0]
        # print( ax.get_proj() )
        projection = ax.get_proj()
        x, y, _ = proj3d.proj_transform( self.el_x, self.el_y, self.el_z, projection )
        xy = np.vstack( [x, y] ).T
        ret = ax.transData.transform( xy )
        print( 'ret.shape=', ret.shape )
        self.x_scr = ret[:,0]
        self.y_scr = ret[:,1]
        if False:
            for i in range(5):
                print( 'point[%d]=' % i, ( self.x[i], self.y[i], self.r[i] ) )
                # print( 'transformed point[%d]=' % i, (x[i], y[i]) )
                print( 'transformed screen point[%d]=' % i, (self.x_scr[i], self.y_scr[i]) )

    def on_button_press( self, event ):
        if event.button == 3:
            self.on_right_button_press( event )
            return

    def on_right_button_press( self, event ):
        print( 'on_right_button_press' )

        if event.inaxes not in self.axes:
            return

        self.compute_projected_xy()

        print( 'x=', event.x )
        print( 'y=', event.y )
        w, h, x, y = split_geometry( self.canvas_frame.mpl_canvas_widget.winfo_geometry() )
        h_ = h
        print( 'self.mpl_canvas_widget.winfo_geometry()=', (w, h, x, y) )

        w, h, x, y = split_geometry( self.dialog.winfo_geometry() )
        # print( 'self.winfo_geometry()=', (w, h, x, y) )
        # TODO: better way to get the mouse cursor position

        elution_no = self.guess_pointed_elution_no( event )
        if elution_no is None:
            return

        self.elution_no = elution_no
        x_  = x if self.plot_no == 0 else x + w//2

        # self.folder_no.set( folder_no )
        self.popup_menu.entryconfig( 0, label="Guinier analysis on %d-th elution" % elution_no )
        self.popup_menu.entryconfig( 1, label="Guinier analysis on %d-th elution corrected" % elution_no )
        self.popup_menu.post( int( x_ + event.x + 10 ), int( y + h_ - event.y + 50 ) )

    def guess_pointed_elution_no( self, event ):
        dist = ( self.x_scr - event.x )**2 + ( self.y_scr - event.y )**2
        i = np.argmin( dist )
        return i

    def update_cursor( self, cursor ):
        for w in [self.canvas_frame.mpl_canvas_widget]:
            w.config( cursor=cursor )

    def guinier_analysis( self, elution_no=None, corrected=False ):
        if elution_no is None:
            elution_no = self.elution_no

        print( 'guinier_analysis: ', elution_no )
        self.update_cursor( 'wait' )

        from SimpleGuinier          import SimpleGuinier
        from molass_legacy.AutorgKekAdapter       import AutorgKekAdapter
        from SimpleGuinierAnimation import SimpleGuinierAnimation
        from CanvasDialog           import CanvasDialog

        if corrected:
            data    = self.corrected_data
        else:
            data    = self.data

        s_data  = data[elution_no,]
        guinier = SimpleGuinier( s_data, anim_data=True )
        adapter = AutorgKekAdapter( None, guinier=guinier )
        adapter_result = adapter.run()

        dialog = CanvasDialog( "SimpleGuinier Animation", parent=self )
        anim_iter_max = 300
        title = 'Guinier analysis of the ' + str(elution_no) + '-th elution'

        def anim_func( fig  ):
            anim = SimpleGuinierAnimation( guinier, title=title, result_rg=adapter_result.Rg  )
            anim.draw( fig, anim_iter_max=anim_iter_max )
            dialog.mpl_canvas.draw()

        self.update_cursor( '' )

        dialog.show( anim_func, figsize=( 18, 5 ) )

class Scattering3dSimpleDialog( Dialog ):
    def __init__( self, parent, serial_data, data_list=None, title_list=None ):
        self.parent = parent
        self.sd     = serial_data
        if data_list is None:
            data_list   = [ serial_data.intensity_array ]
        self.data_list  = data_list
        if title_list is None:
            title_list  = []

            title = 'Xray Scattering from ' + get_setting( 'in_folder' )
            for n in range( len(self.data_list) ):
                sim_title = '' if n == 0 else 'Guinier-Porod-fitted '
                title_list.append( sim_title + title )

        self.title_list = title_list
        self.applied    = False

    def show( self ):
        title = "Scattering3dSimplePlot"
        Dialog.__init__(self, self.parent, title )

    def body( self, body_frame ):
        tk_set_icon_portable( self )

        plot_frame = Tk.Frame( body_frame )
        plot_frame.pack()

        self.sync_view = Tk.IntVar()
        self.sync_view.set(0)

        self.plot_frames = []

        for i, data in enumerate(self.data_list):
            # add ", sim_base=self.sim_base" when plotting simulated base
            plot3d_frame = Scattering3dSimpleFrame( plot_frame, self, self.sd, data, self.title_list[i], i )
            plot3d_frame.grid( row=0, column=i )
            self.plot_frames.append( plot3d_frame )

        if len(self.plot_frames) > 1:
            self.sync_view.set(1)

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        frame = self

        box2 = Tk.Frame( frame )
        box2.pack( fill=Tk.X, expand=1 )

        frames = []
        for k in range(3):
            f = Tk.Frame( box2 )
            f.pack( side=Tk.LEFT, fill=Tk.X, expand=1 )
            frames.append( f )

        w = Tk.Button(frames[2], text="Cancel", width=10, command=self.cancel, default=Tk.ACTIVE)
        w.pack(side=Tk.RIGHT, padx=10, pady=10)
        self.cancel_button = w

        w = Tk.Button(frames[2], text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.RIGHT, padx=10, pady=10)
        self.ok_button = w

        self.apply_fitted_base = Tk.IntVar()
        self.apply_fitted_base.set(0)

        if len(self.plot_frames) > 1:
            w = Tk.Checkbutton(frames[2], text="Apply Fitted Correction", variable=self.apply_fitted_base )
            w.pack(side=Tk.RIGHT, padx=10, pady=10)

            cb = Tk.Checkbutton( frames[1], text="Synchronize view", variable=self.sync_view  )
            cb.pack( side=Tk.RIGHT, padx=20 )

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def apply( self ):  # overrides parent class method
        self.applied        = True



"""
    Models.Stochastic.OligoporeDistribution.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from datetime import datetime
from importlib import reload

BH_NITER = 100

def estimate_oligopore_distribution_separately(moment_rg_file, column_type, debug=False):
    from molass_legacy.Batch.LiteBatch import LiteBatch
    from molass_legacy.Models.Stochastic.DatablockUtils import load_datablock_list
    from molass_legacy.Models.Stochastic.OligoporePlotImpl import get_block_lrf_src
    from molass_legacy.Models.Stochastic.OligoporeMoments import study_oligopore_moments_impl

    in_folders, block_list = load_datablock_list(moment_rg_file, column_type)
    batch = LiteBatch()
    params_list = []
    for in_folder, block in zip(in_folders, block_list):
        lrf_src = get_block_lrf_src(batch, in_folder)
        N, K, t0, pszs, pszp = study_oligopore_moments_impl(lrf_src, return_guess_params=True)
        colparams = np.concatenate([[N, K, t0], pszs, pszp])
        params_list.append(colparams)
    return params_list

def estimate_oligopore_distribution(moment_rg_file, column_type, params_list=None, rg_curves=None, debug=False):
    from molass_legacy.Models.Stochastic.DatablockUtils import load_datablock_list
    from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments, to_moderate_props
    from molass_legacy.Models.Stochastic.RgReliability import determine_unreliables
    from molass_legacy.Models.Stochastic.ParamLimits import (USE_K, M1_WEIGHT, M2_WEIGHT,
                                               MNP_BOUNDS, PORESIZE_BOUNDS, PROP_BOUND, TIMESCALE_BOUND,
                                               BASINHOPPING_SCALE, N0_INIT, ParamsScaler)
    if debug:
        import Models.Stochastic.OligoporeUtils as module
        reload(module)
    from molass_legacy.Models.Stochastic.OligoporeUtils import guess_oligopore_colparams

    print('estimate_oligopore_distribution')

    assert USE_K

    T0_BOUND = MNP_BOUNDS[2].copy()

    class MultidatasetParamsScaler(ParamsScaler):
        def __init__(self, num_pszs, num_experiments, num_timescales=0):
            self.params_list = params_list
            MNP_BOUNDS_ = MNP_BOUNDS.copy()
    
            bounds_list = (MNP_BOUNDS_[0:2]
                            + [PORESIZE_BOUNDS]*num_pszs      # allowing the upper bound to reach MAX_PORESIZE can cause problems?
                            + [PROP_BOUND]*(num_pszs-1)
                            + [T0_BOUND] * num_experiments
                            + [TIMESCALE_BOUND]*num_timescales)

            self.bounds = np.array(bounds_list)
            self.set_scales()

    in_folders, block_list = load_datablock_list(moment_rg_file, column_type)

    use_basinhopping = True
    use_timescales = True
    num_pszs = 2
    num_blocks = len(block_list)
    sep1 = 2 + num_pszs
    sep2 = sep1 + num_pszs -1
    sep3 = sep2 + num_blocks
    def split(p):
        N, K = p[0:2]
        pszv = p[2:sep1]
        pszp = np.zeros(num_pszs)
        pszp[:-1] = p[sep1:sep2]
        pszp[-1] = 1 - np.sum(pszp[:-1])
        t0v = p[sep2:sep3]
        if use_timescales:
            tscales = p[sep3:]
        else:
            tscales = np.array([])
        return N, K/N, pszv, pszp, t0v, tscales

    num_timescales = num_blocks if use_timescales else 0
    scaler = MultidatasetParamsScaler(num_pszs, num_blocks, num_timescales=num_timescales)

    me = 1.5
    mp = 1.5

    def multi_dataset_objective(p):
        if use_basinhopping:
            p = scaler.scale_back(p)
        N, T, pszv, pszp, t0v, tscales = split(p)

        t0_upper_penalty = 0
        dev = 0
        for k, (block, t0) in enumerate(zip(block_list, t0v)):
            if use_timescales:
                T_ = T*tscales[k]
                t0_ = t0*tscales[k]
            else:
                T_ = T
                t0_ = t0
            for M1, M2, rg, quality in block:
                t0_upper_penalty += min(0, M1 - np.sqrt(M2) - t0_)**2    # i.e., M1 - np.sqrt(M2) > t0 must always hold
                rhov = rg/pszv
                rhov[rhov > 1] = 1
                M1_ = t0_ + N * T_ * np.sum(pszp * (1 - rhov)**(me + mp))
                M2_ = 2 * N * T_**2 * np.sum(pszp * (1 - rhov)**(me + 2*mp))
                dev += quality * np.log(M1_WEIGHT*(M1_ - M1)**2 + M2_WEIGHT*(M2_ - M2)**2)
                # dev += np.log(M1_WEIGHT*(M1_ - M1)**2 + M2_WEIGHT*(M2_ - M2)**2)

        return dev + t0_upper_penalty*1e6

    if use_timescales:
        if column_type == '1':
            timescales = [1, 1, 1, 1, 1]
        else:
            timescales = [1, 1, 1, 0.6, 0.6]
    else:
        timescales = []

    if params_list is None:
        init_t0v = []
        for i, block in enumerate(block_list):
            t0_init = np.min([M[0] - np.sqrt(M[1])*20 for M in block[:,1:3]])
            init_t0v.append(t0_init)

        t0_upper = np.max(block[:,1] - np.sqrt(block[:,2]))     # 

        print("init_t0v=", init_t0v)
        init_params = np.concatenate([[1000, 1000], [60, 200], [0.5], init_t0v, timescales])
        bounds = [(300,2000)]*2 + [(30,500)]*num_pszs + [(0,1)]*(num_pszs - 1) + [(-1000, t0_upper)]*num_blocks + [TIMESCALE_BOUND]*num_timescales
    else:
        params_array = np.array(params_list)
        # print("params_array=", params_array)
        init_t0v = params_array[:,2]
        weighted_array = compute_weighted_array(params_array)
        weighted_avarage = np.average(weighted_array, axis=0)
        N, K = weighted_avarage[0:2]
        init_pszv = weighted_avarage[3:3+num_pszs]
        init_pszp = np.average(params_array[:,3+num_pszs:3+2*num_pszs-1], axis=0)
        print("weighted_avarage=", weighted_avarage)
        print("init_pszv=", init_pszv)
        print("init_pszp=", init_pszp)
        print("init_t0v=", init_t0v)
        init_params = np.concatenate([[N, K], init_pszv, init_pszp, init_t0v, timescales])
        t0_upper = np.max(init_t0v)
        bounds = [(300,2000)]*2 + [(30,500)]*num_pszs + [(0,1)]*(num_pszs - 1) + [(-1000, t0_upper)]*num_blocks + [TIMESCALE_BOUND]*num_timescales

    if use_basinhopping:
        print("init_params=", init_params)
        print("bounds=", bounds)
        bounds_ = [(0, BASINHOPPING_SCALE)] * len(init_params)
        minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds_)
        minima_counter = 0
        def minima_callback(x, f, accept):
            nonlocal minima_counter
            if minima_counter % 10 == 0:
                print("%s: minima_counter=%d" % (datetime.now().strftime("%Y-%m-%d %H:%M:%S"), minima_counter))
            minima_counter += 1

        init_params_ = scaler.scale(init_params)

        res = basinhopping(multi_dataset_objective, init_params_, niter=BH_NITER,
                           minimizer_kwargs=minimizer_kwargs, callback=minima_callback)
        res.x = scaler.scale_back(res.x)
    else:
        res = minimize(multi_dataset_objective, init_params, method='Nelder-Mead', bounds=bounds)
    print('res.x=', res.x)

    N, T, pszv, pszp, t0v, tscales = split(res.x)

    import molass_legacy.KekLib.DebugPlot as plt

    extra_button_specs=[("Plot Results", lambda : plot_results(N, T, pszv, pszp, t0v, tscales, in_folders, block_list, params_list, rg_curves=rg_curves)),]
    with plt.Dp(button_spec=["OK", "Cancel"], extra_button_specs=extra_button_specs):
        fig, ax = plt.subplots()
        plt.show()

def rounded_list(fmt, values):
    return ', '.join([fmt % v for v in values])

def plot_results(N, T, pszv, pszp, t0v, tscales, in_folders, block_list, params_list, rg_curves=None):
    import Models.Stochastic.OligoporePlotImpl as module
    reload(module)
    from molass_legacy.Models.Stochastic.OligoporePlotImpl import plot_results_impl
    try:
        plot_results_impl(N, T, pszv, pszp, t0v, tscales, in_folders, block_list, params_list, rg_curves=rg_curves)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None, "plot_results_impl failure: ")

def load_csv_values(path):
    params_list = []
    with open(path) as fh:
        for line in fh:
            values = [float(v) for v in line[:-1].split(',')]
            params_list.append(values)
    return params_list

def compute_weighted_array(params_array):
    pv = params_array[:,-2]
    qv = params_array[:,-1]
    target_array = params_array[:,:-2].T
    return (target_array*pv + target_array*qv).T

def test_weighted_array():
    from molass_legacy.SerialAnalyzer.DataUtils import get_local_path
    TODO = get_local_path('TODO')
    params_array = np.array(load_csv_values(TODO + r'\20240425\params_list.csv'))
    print("params_array=", params_array)
    print("weighted_average=", compute_weighted_array(params_array))

if __name__ == "__main__":
    import sys
    sys.path.append("../lib")
    import os
    import seaborn as sns
    sns.set_theme()
    from molass_legacy.KekLib.ChangeableLogger import Logger
    from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry
    from molass_legacy.SerialAnalyzer.DataUtils import get_local_path
    from molass_legacy.Models.Stochastic.DatablockUtils import get_rg_curves

    TODO = get_local_path('TODO')
    column_type = '1'

    # test_weighted_array()
    # exit()

    # os.makedirs('temp', exist_ok=True)
    clear_dirs_with_retry(['temp'])
    logger = Logger("temp/psd-estimate.log")

    moment_rg_file = TODO + r'\20240425\rgs.csv'
    if False:    
        params_list = estimate_oligopore_distribution_separately(moment_rg_file, column_type)

        with open('temp/params_list-%s.csv' % column_type, 'w') as fh:
            for params in params_list:
                fh.write(','.join(['%g' % v for v in params]) + '\n')
    else:
        rg_folders_root = TODO + '\\20240523\\rg_folders'
        rg_curves = get_rg_curves(rg_folders_root, column_type)
        params_list = load_csv_values(TODO + r'\20240425\params_list-%s.csv' % column_type)
        estimate_oligopore_distribution(moment_rg_file, column_type, params_list=params_list, rg_curves=rg_curves)


# coding: utf-8
"""
    QmmDialog.py

    Copyright (c) 2020-2024, SAXS Team, KEK-PF
"""
import os
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon
import matplotlib.ticker as ticker
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar, get_color
import molass_legacy.KekLib.DebugPlot as dplt
from OurPlotUtils import draw_as_image
from DataUtils import get_in_folder
from .GroupingMatrix import GroupingMatrix, num_to_char
from molass_legacy.AutorgKekAdapter import AutorgKekAdapter
from molass_legacy._MOLASS.SerialSettings import get_setting
from .QmmDenssMenu import QmmDenssMenu

DISTINCTION_Q = 0.15
DISTINCTION_WL = 220
APPLY_DENSS_FIT = True
if APPLY_DENSS_FIT:
    from molass_legacy.DENSS.DenssUtils import fit_data

class QmmDialog(Dialog):
    def __init__(self, parent, controller, **kwargs):
        self.applied = False
        self.datasets = controller.datasets
        self.qmm = controller.qmm
        self.model_name = self.qmm.get_model_name()
        self.y_list = controller.y_list
        self.data_list = controller.data_list
        cd = get_setting('conc_dependence')
        self.consider_ipe = cd == 2
        Dialog.__init__(self, parent, title="%s Quad Mixture Model" % self.model_name, visible=False)

    def show(self):
        self._show()

    def body(self, bframe):
        self.base_frames = []
        for k, dtype in enumerate(["UV", "Xray"]):
            k2 = k*2
            k_slice = slice(k2, k2+2)
            frame = QmmFrame(bframe, self, self.datasets, dtype, self.qmm,
                                self.y_list[k_slice])
            frame.pack()
            self.base_frames.append(frame)

        self.current_frame = 1
        self.base_frames[1 - self.current_frame].pack_forget()

    def get_current_frame(self):
        return self.base_frames[self.current_frame]

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, expand=1, padx=10)

        for j in range(3):
            box.columnconfigure(j, weight=1)

        w = Tk.Button(box, text="◀ Main", width=10, command=self.cancel)
        w.grid(row=0, column=0, sticky=Tk.W, padx=10, pady=5)

        sw_text = self.get_switch_text()
        w = Tk.Button(box, text=sw_text, command=self.toggle)
        w.grid(row=0, column=1)
        self.switch_button = w

        w = Tk.Button(box, text="▶ Serial Analysis", command=self.ok, state=Tk.DISABLED)
        w.grid(row=0, column=2, sticky=Tk.E, padx=10, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def get_switch_text(self):
        dtype = 'UV' if self.current_frame == 1 else 'Xray'
        return "◇ Switch to " + dtype

    def toggle(self):
        i = self.current_frame
        j = 1 - i
        self.base_frames[i].pack_forget()
        self.base_frames[j].pack()
        self.current_frame = j
        sw_text = self.get_switch_text()
        self.switch_button.config(text=sw_text)

    def apply(self):
        self.applied = True

    def get_rg_array(self):
        return self.base_frames[1].get_rg_array()

class QmmFrame(Tk.Frame):
    def __init__(self, parent, dialog, datasets, dtype, qmm, y_list):
        Tk.Frame.__init__(self, parent)
        self.dialog = dialog
        self.model_name = dialog.model_name
        self.datasets = datasets
        self.S_init = None
        self.dtype = dtype
        self.rg_set_list = None
        self.i_pos_values = datasets.get_i_pos_values(dtype)
        self.i_pos_format = '@WL=%.3g' if dtype == 'UV' else '@Q=%.2g'
        if dtype=='UV':
            self.center_title = "Component Absorbance Curves"
            self.no_base = 0
            self.dataset = self.datasets.pair[0]
            self.pos_format = '@WL=%.3g'
            self.pos_list = [DISTINCTION_WL]
            self.consider_ipe = False
        else:
            self.logger = logging.getLogger(__name__)
            self.center_title = "Component Scattering Curves"
            self.no_base = 2
            self.dataset = self.datasets.pair[1]
            self.pos_format = '@Q=%.3g'
            self.pos_list = [DISTINCTION_Q]
            self.consider_ipe = dialog.consider_ipe

        self.j0 = self.dataset.j0
        main_frame = Tk.Frame(self)
        main_frame.pack()
        bottom_frame = Tk.Frame(self)
        bottom_frame.pack(fill=Tk.X, expand=1)
        left_frame = Tk.Frame(main_frame)
        left_frame.pack(side=Tk.LEFT)
        right_frame = Tk.Frame(main_frame)
        right_frame.pack(side=Tk.LEFT)
        toolbar_frame = Tk.Frame(bottom_frame)
        toolbar_frame.pack(side=Tk.LEFT)
        button_frame = Tk.Frame(bottom_frame)
        button_frame.pack(side=Tk.RIGHT)

        self.build_canvases(left_frame, dtype, qmm, y_list, toolbar_frame)
        self.build_panels(right_frame, button_frame)
        self.update()
        self.copy_to_panel_fig(self.fig, self.axes[0,:])
        self.popup_menu = None

    def on_mpl_button_press(self, event):
        if event.xdata is None:
            return

        if event.button == 3:
            from molass_legacy.KekLib.TkUtils import split_geometry
            self.create_popup_menu()
            w, h, x, y = split_geometry(self.dialog.geometry())
            self.popup_menu.post(x + event.x + 20, y + h - event.y - 50)
            return

    def create_popup_menu(self):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0)
            self.popup_menu.add_command(label='Show 3D Figure', command=self.show_3d_figure)
            self.popup_menu.add_command(label='Vaious Denoise Figures', command=self.show_various_denoise_figures)

    def grouped(self, S):
        if self.S_init is None or S is None:
            return False

        sdiff = np.sum(np.abs(S - self.S_init))
        return sdiff > 0

    def build_canvases(self, frame, dtype, qmm, y_list, toolbar_frame):
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(18, 10))
        self.fig = fig
        self.axes = axes
        if self.dtype == 'Xray':
            self.twin_axes = [ax.twinx() for ax in axes[:,2]]

        self.mpl_canvas = FigureCanvasTkAgg(fig, frame)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        if self.dtype == 'Xray':
            self.mpl_canvas.mpl_connect('button_press_event', self.on_mpl_button_press)

        fig.suptitle("Decomposition of %s Data from %s using %s-QMM" % (dtype,  get_in_folder(), self.model_name), fontsize=30)

        xy_list, C_list, gy_list, mu_list, sigma_list = self.get_draw_lists(qmm, y_list)
        self.num_components = qmm.K
        self.mu_list = mu_list
        self.sigma_list = sigma_list
        self.draw_all(xy_list, C_list, gy_list)
        self.xy_list = xy_list
        self.C_list = C_list
        self.gy_list = gy_list

        fig.tight_layout()
        fig.subplots_adjust(top=0.9)
        self.modify_major_xtick_positions(shrink_axis=True)

        self.mpl_canvas.draw()
        self.toolbar = NavigationToolbar(self.mpl_canvas, toolbar_frame)
        self.toolbar.update()

    def get_draw_lists(self, qmm, y_list):
        xy_list = []
        C_list = []
        gy_list = []
        mu_list = []
        sigma_list = []
        for n, y in enumerate(y_list):
            mm = qmm.get_sub(self.no_base+n)
            x = np.arange(len(y))
            C, gy = mm.get_anim_C(x, y, -1, total=True)
            xy_list.append((x,y))
            C_list.append(C)
            gy_list.append(gy)
            mu_list.append(mm.get_peak_mean_x())
            sigma_list.append(mm.sigma)
        return xy_list, C_list, gy_list, mu_list, sigma_list

    def draw_all(self, xy_list, C_list, gy_list, k_rows_list=None, effective_sizes=None):
        axes = self.axes
        self.draw_elution(axes[:,0], xy_list, C_list, gy_list, effective_sizes)
        ret_list = self.draw_featuring_curves(axes[:,1], C_list, effective_sizes)
        point_set_list = self.draw_distinction(axes[:,2], xy_list, ret_list, k_rows_list)
        if k_rows_list is None:
            self.point_set_list = np.array(point_set_list)

    def draw_elution(self, axes, xy_list, C_list, gy_list, effective_sizes=None):
        for n, (ax, xy, C, gy) in enumerate(zip(axes, xy_list, C_list, gy_list)):
            ax.cla()
            if n == 0:
                ax.set_title("Component Elution Curves", fontsize=16)
            x, y = xy
            x_ = x + self.j0
            ax.plot(x_, y, label='input')
            ax.plot(x_, gy, label='total')
            size = None if effective_sizes is None else effective_sizes[n]
            self.plot_components(ax, x_, C[:size])
            self.draw_i_pos_text(n, ax)
            ax.legend()

    def plot_components(self, ax, x, cgys):
        ret_lines = []
        for k, gy in enumerate(cgys):
            line, = ax.plot(x, gy, ':', label='c-%s' % num_to_char(k))
            ret_lines.append(line)
        return ret_lines

    def draw_i_pos_text(self, n, ax, fontsize=30, xpos=0.2):
        text = self.i_pos_format % self.i_pos_values[n]
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        tx = xmin*(1-xpos) + xmax * xpos
        ty = ymin*0.1 + ymax * 0.9
        ax.text(tx, ty, text, ha='center', va='center', fontsize=fontsize, alpha=0.1)

    def draw_featuring_curves(self, axes, C_list, effective_sizes=None):
        self.data_list = []     # for DENSS
        self.result_list = []
        ret_list = []
        for n, (ax, C) in enumerate(zip(axes, C_list)):
            ax.cla()
            if n == 0:
                ax.set_title(self.center_title, fontsize=16)
            size = None if effective_sizes is None else effective_sizes[n]
            _, ylist, e11n = self.datasets.draw_exprapolated(self.no_base+n, ax, C, size, ipe=self.consider_ipe)
            self.result_list.append(e11n)
            ret_list.append(ylist)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for p in [self.i_pos_values[n], self.pos_list[0]]:
                ax.plot([p, p], [ymin, ymax], ':', color='gray')
            self.draw_i_pos_text(n, ax, xpos=0.5)
        return ret_list

    def draw_distinction(self, axes, xy_list, ret_list, k_rows_list=None):
        point_set_list = []
        rg_set_list = []

        for n, (ax, xy, ylist) in enumerate(zip(axes, xy_list, ret_list)):
            x = self.mu_list[n] + self.j0

            e11n = self.result_list[n]
            ax.cla()

            if self.dtype == 'Xray':
                axt = self.twin_axes[n]
                axt.cla()
                axt.grid(False)
            if n == 0:
                ax.set_title("Components Distinction", fontsize=16)

            if k_rows_list:
                # print('self.point_set_list[%d]=' % n, self.point_set_list[n])
                for k, y_ in self.point_set_list[n]:
                    ax.plot(x[int(k)], y_, 'o', color='gray', alpha=0.3)

            if k_rows_list:
                k_rows = k_rows_list[n]
            else:
                k_rows = None

            assert len(self.pos_list) == 1

            pos_text = self.pos_format % self.pos_list[0]

            for p in self.pos_list:
                i = self.dataset.get_row_index(p)
                # picked_line = np.average(ylist[:,i-2:i+3], axis=1)    # this won't necessarily impove the line's stability
                picked_line = ylist[:,i]
                if k_rows is None:
                    y_for_line = picked_line
                else:
                    init_y_list = self.point_set_list[n,:,1]
                    y_for_line = self.expand_line(init_y_list, k_rows, picked_line)
                ax.plot(x, y_for_line, ':', label='Intensities' + pos_text)
                point_set = []
                for k, y_ in enumerate(picked_line):
                    if k_rows is None:
                        ax.plot(x[k], y_, 'o', color='C%d' % (k+2))
                        point_set.append((k, y_))
                    else:
                        k_ = k_rows[k]
                        y_ = np.ones(len(k_))*y_
                        ax.plot(x[k_], y_, marker='o', linewidth=3, color='C%d' % (k+2))

                if k_rows is None:
                    point_set_list.append(point_set)

            if self.dtype == 'Xray':
                rg_array = self.compute_rg_list(e11n)
                if k_rows is None:
                    rg_line = rg_array
                    rg_set_list.append(rg_array)
                else:
                    init_rg_array = self.rg_set_list[n]
                    rg_line = self.expand_line(init_rg_array, k_rows, rg_array)
                rg_line_y = rg_line[:,0]
                try:
                    axt.errorbar(x, rg_line_y, rg_line[:,1], fmt='o:', markersize=5, capsize=3, elinewidth=3, mfc='black', label="Rg's")
                except:
                    from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                    etb = ExceptionTracebacker()
                    self.logger.warning("resorting to plain plot because errorbar plot failed due to %s", etb.last_lines())
                    axt.plot(x, rg_line_y, ':', marker='^', label="Rg's")
                quality = rg_line[:,2]
                for quality_limit in [0.5, 0.3]:
                    is_major = np.logical_and(np.isfinite(rg_line_y), quality > quality_limit)
                    major_rgs = rg_line_y[is_major]
                    major_average = np.average(major_rgs)
                    if len(major_rgs) > 2 and np.isfinite(major_average):
                        break
                if not np.isnan(major_average):
                    if len(major_rgs) > 1:
                        major_std = np.std(major_rgs)
                        print('major_rgs=', major_rgs, 'major_std=', major_std)
                        dy = max(10, major_std*2)
                    else:
                        dy = 10
                    axt.set_ylim(major_average-dy, major_average+dy)

                if False:
                    self.add_rg_values_from_v1_result(axt)

                # axt.legend(loc='center right')
                axt.legend(bbox_to_anchor=(1, 0.92), loc='upper right')

            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = xmin*0.8 + xmax * 0.2
            ty = ymin*0.1 + ymax * 0.9
            ax.text(tx, ty, pos_text, ha='center', va='center', fontsize=30, alpha=0.1)

            ax0 = self.axes[n,0]
            ax.set_xlim(ax0.get_xlim())
            ymin0, ymax0 = ax0.get_ylim()

            def transform_y(y):
                return (y - ymin0)*(ymax - ymin)/(ymax0 - ymin0) + ymin

            xp, yp = xy
            xp_ = xp + self.j0
            yp_ = transform_y(yp)
            zp_ = transform_y(0)
            poly_points = [(xp_[0], zp_)] + list(zip(xp_, yp_)) + [(xp_[-1], zp_)]
            poly = Polygon( poly_points, alpha=0.1, color='gray')
            ax.add_patch(poly)

            ax.legend()
            if self.rg_set_list is None:
                self.rg_set_list = rg_set_list

        return point_set_list

    def modify_major_xtick_positions(self, shrink_axis=False):
        # should be called after subplots_adjust
        for n, ax in enumerate(self.axes[:,2]):
            if shrink_axis:
                box = ax.get_position()
                dh = box.height * 0.05
                ax.set_position([box.x0, box.y0+dh, box.width, box.height-dh])

            ax0 = self.axes[n,0]
            ax.xaxis.set_major_locator(ticker.FixedLocator(ax0.get_xticks()))
            for k, label in enumerate(ax.xaxis.get_majorticklabels()):
                label.set_y(-0.05)
            ax.grid(False, which='major')

            def format_func(value, tick_number):
                """
                Customizing Ticks
                https://jakevdp.github.io/PythonDataScienceHandbook/04.10-customizing-ticks.html
                """
                return num_to_char(tick_number)

            x = self.mu_list[n] + self.j0
            ax.xaxis.set_minor_locator(ticker.FixedLocator(x))
            ax.xaxis.set_minor_formatter(ticker.FuncFormatter(format_func))
            ax.grid(True, which='minor')

    def compute_rg_list(self, e11n):
        denss_fitted_rg = get_setting('denss_fitted_rg')
        rg_list = []
        q = self.dataset.vector
        P = e11n.P
        E = e11n.E
        num_curves = P.shape[1]//2 if self.consider_ipe else P.shape[1]
        for k in range(num_curves):
            y_ = P[:,k]
            e_ = E[:,k]

            if APPLY_DENSS_FIT and denss_fitted_rg:
                q_, y_, e_, _ = fit_data(q, y_, e_)
            else:
                q_ = q

            A_data = np.vstack( [q_, y_, e_] ).T
            autorg_kek = AutorgKekAdapter( A_data )
            result = autorg_kek.run()
            if result.Rg is None:
                rg_list.append((np.nan, np.nan, np.nan))
            else:
                rg_list.append((result.Rg, result.Rg_stdev, result.Quality))

        self.rg_array = np.array(rg_list)
        return self.rg_array

    def get_rg_array(self):
        return self.rg_array

    def expand_line(self, init_y_list, k_rows, picked_line):
        ret_y_list = init_y_list.copy()
        for k_row, y in zip(k_rows, picked_line):
            ret_y_list[k_row] = y
        return ret_y_list

    def build_panels(self, frame, button_frame):
        fig_frame = Tk.Frame(frame)
        fig_frame.pack(padx=10)
        space = Tk.Frame(frame, height=5)
        space.pack()
        matrix_frame = Tk.Frame(frame)
        matrix_frame.pack()

        fig, axes = plt.subplots(nrows=3, ncols=1, figsize=(3, 7.2))
        self.fig2 = fig
        self.axes2 = axes
        self.mpl_canvas2 = FigureCanvasTkAgg(fig, fig_frame)
        self.mpl_canvas2_widget = self.mpl_canvas2.get_tk_widget()
        self.mpl_canvas2_widget.pack(fill=Tk.BOTH, expand=1)

        label = Tk.Label(matrix_frame, text="Reduction Matrix", font=Tk.font.Font(size=20))
        label.pack()
        k = self.num_components
        self.gm = GroupingMatrix(matrix_frame, nrows=k, ncols=k, read_only=False)
        self.gm.pack()
        self.S_init = self.gm.get_matrix().copy()

        reset_button = Tk.Button(button_frame, text="Reset", width=10, command=self.reset)
        reset_button.pack(side=Tk.LEFT, padx=5, pady=5)
        compute_button = Tk.Button(button_frame, text="Compute", width=10, command=self.compute)
        compute_button.pack(side=Tk.LEFT, padx=5, pady=5)
        save_button = Tk.Button(button_frame, text="Save", width=10, command=self.save)
        save_button.pack(side=Tk.LEFT, padx=5, pady=5)

        if self.dtype == 'Xray':
            self.denss_menu = QmmDenssMenu(button_frame, self, state=Tk.DISABLED)
            self.denss_menu.pack(side=Tk.LEFT, padx=5)

    def copy_to_panel_fig(self, from_fig, from_axes):
        for to_ax, from_ax in zip(self.axes2, from_axes):
            to_ax.set_axis_off()
            to_title = from_ax.get_title().replace('Component', 'Individual')
            to_ax.set_title(to_title, y=0.98)
            draw_as_image(to_ax, from_fig, from_ax)
        self.fig2.subplots_adjust(top=0.95, bottom=0, left=0, right=1, wspace=0)
        self.mpl_canvas2.draw()

    def reset(self):
        self.gm.reset()
        self.compute()

    def compute(self):
        S = self.gm.get_matrix()
        assert S.shape[0] == S.shape[1]

        grouped_C_list, k_rows_list, effective_sizes = self.get_grouped_C_list(S)
        self.draw_all(self.xy_list, grouped_C_list, self.gy_list, k_rows_list, effective_sizes)
        self.modify_major_xtick_positions()
        self.mpl_canvas.draw()

        if self.dtype == 'Xray':
            self.denss_menu.config(state=Tk.NORMAL if self.grouped(S) else Tk.DISABLED)

    def get_grouped_C_list(self, S):
        new_C_list = []
        effective_sizes = []
        k_rows_list = []
        for C in self.C_list:
            C_rows = []
            k_rows = []
            for i in range(S.shape[0]):
                g_list = []
                k_vect = []
                for j, selected in enumerate(S[i,:]):
                    if selected:
                        g_list.append(C[j,:])
                        k_vect.append(j)
                if len(g_list) > 0:
                    c = np.sum(g_list, axis=0)
                    C_rows.append(c)
                    k_rows.append(np.array(k_vect))
            k_rows_list.append(k_rows)
            effective_sizes.append(len(C_rows))
            for j in range(S.shape[1]):
                if np.sum(S[:,j]) == 0:
                    # output-suppressed C_row,
                    # which will be exluded using effective_sizes
                    C_rows.append(C[j,:])
            new_C_list.append(np.array(C_rows))

        ret_k_rows_list = k_rows_list if self.grouped(S) else None
        return new_C_list, ret_k_rows_list, effective_sizes

    def save(self):
        from .QmmResultSaver import QmmResultSaverDialog
        dialog = QmmResultSaverDialog(self.dialog, self.result_list[0])
        dialog.show()

    def save_the_figure(self, folder, analysis_name):
        filename = analysis_name.replace( 'analysis', 'figure' )
        path = os.path.join( folder, filename )
        self.fig.savefig( path )
        return path

    def show_3d_figure(self):
        from .Qmm3dFigure import Qmm3dFigure
        fig = Qmm3dFigure(self.dialog, self.dataset, self)
        fig.show()

    def show_various_denoise_figures(self):
        from .DenoiseFigures import DenoiseFigures
        fig = DenoiseFigures(self.dialog, self.dataset, self)
        fig.show()

    def add_rg_values_from_v1_result(self, ax):
        from DataUtils import get_pytools_folder
        from molass_legacy.Reports.ReportUtils import just_load_guinier_result
        from molass_legacy.KekLib.NumpyUtils import np_loadtxt
        report_folder = get_pytools_folder() + '/reports/20180602'
        # report_folder = get_pytools_folder() + '/reports/sample_data'
        guinier_result_csv = report_folder + '/.guinier_result/--serial_result.csv'
        array = np.array(just_load_guinier_result(guinier_result_csv))
        self.logger.info('array.shape=%s', str(array.shape))
        xr = self.dataset
        j0 = xr.j0
        # j0= 604 xr.data.shape= (1074, 101) array.shape= (311, 33)
        print('j0=', j0, 'xr.data.shape=', xr.data.shape, 'array.shape=', array.shape)
        size = xr.data.shape[1]
        x = j0 + np.arange(size)
        y = np.array([float(rg) for rg in array[:,11]])
        ax.plot(x, y[j0:j0+size], color='green', label="Rg's (V1 Guinier)")

        extrapolated_csv = report_folder + '/extrapolated.csv'
        xy, _ = np_loadtxt(extrapolated_csv)
        ax.plot(xy[:,0], xy[:,1], color='orange', label="Rg's (V1 Extrapolated)")



"""
    ModelEvaluator.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from molass_legacy.SerialAnalyzer.ElutionCurve import ANALYSIS_RANGE_RATIO

class ModelEvaluator:
    def __init__(self, model, params, sign=1, accepts_real_x=False, debug=False):
        self.model = model
        self.model_name = model.get_name()
        if model.is_traditional():
            self.param_values = params[0:5].copy()  # why 0:5
        else:
            self.param_values = params.copy()

        if len(params) > 5:
            # ('const_a', <Parameter 'const_a', value=14.635374905168955 +/- 0.565, bounds=[0:inf], expr='sigma*1.5 - abs(a)'>)])
            # self.const_a = params[5]
            pass

        self.sign = sign
        self.accepts_real_x = accepts_real_x    # for temporary fix under V1PreviewAdapter
        self.func = model.func

    def get_model_name( self ):
        return self.model_name

    def get_func(self):
        return self.func

    def set_new_params(self, param_values):
        self.param_values = param_values

    def get_param_value(self, k):
        return self.param_values[k]

    def get_param_values(self):
        return self.param_values[0:4]

    def get_all_param_values(self):
        return self.param_values

    def get_all_params_string(self):
        values = self.param_values.copy()
        values[0] *= self.sign      # TODO: include sign to h instead of multiplying here
        return self.model.get_params_string(values)

    def get_range_params( self, x, ratio=ANALYSIS_RANGE_RATIO ):
        """
        TODO: consider unifying this logic with EmgPeak.get_model_x_from_ratio(...)
        """
        y = self.__call__(x)
        top_x = np.argmax(y)
        top_y = y[top_x]
        foot_y = top_y * ratio
        i = bisect_right(y[0:top_x], foot_y)
        y_ = y[top_x:]
        j = bisect_right(y_[::-1], foot_y)
        return i, top_x, len(x) - 1 - j

    def get_model_def_expr( self ):
        return self.model.get_def_expr(self.lmfit_params)

    def update_param( self, k, value ):
        self.param_values[k] = value

    def __call__(self, x, debug=False):
        # 
        if debug:
            print("ModelEvaluator.__call__: param_values==", self.param_values)
        if np.isscalar(x):
            return self.func(np.array([x]), *self.param_values)[0] * self.sign
        else:
            return self.func(x, *self.param_values) * self.sign

    def __repr__( self ):
        return "%s(%s(),%s)" % (self.__class__.__qualname__, self.get_model_name(), repr(self.param_values))

    def x_from_height_ratio(self, ecurve, ratio):
        return self.model.x_from_height_ratio(ecurve, ratio, self.param_values)



"""
    PeakInfo.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""

class PeakInfo:
    def __init__(self, peak_id, top_x, elm_recs=None):
        self.peak_id    = peak_id
        self.top_x      = top_x

        self.elm_recs   = elm_recs
        """
        self.elm_recs will be
            either
                None
                    when constructed from old-style range
                    in AnalysisRangeInfo.upgrade_ranges
            or
                not None
                    when constructed
                    in DecompUtils.make_range_info_impl
        """

    def __repr__(self):
        elm_recs_ = None if self.elm_recs is None else ''.join([ str(e) for e in self.elm_recs ])
        return "%s(%d,%g,%s)" % (self.__class__.__qualname__, self.peak_id, self.top_x, elm_recs_)


# coding: utf-8
"""
    InputData.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import logging
import time
import numpy as np
from OurProfiler import take_time

class InputData:
    def __init__(self, in_folder, teller=None, debug=False):
        from SerialDataUtils import get_uv_filename, get_mtd_filename

        self.logger = logging.getLogger(__name__)
        self.in_folder = in_folder
        self.teller = teller
        self.debug = debug
        self.for_sec = True
        self.xray_only = False
        filename = get_uv_filename(in_folder)
        if filename is None:
            filename = get_mtd_filename(in_folder)
            if filename is None:
                self.xray_only = True
            else:
                self.for_sec = False

        self.filename = filename
        if self.teller is None:
            self.load_data(self.in_folder)
        else:
            self.start_prepare()

    def start_prepare(self):
        self.teller.start_load_data(self.in_folder, self.for_sec, self.xray_only)

    def wait_for_ready(self):
        teller = self.teller
        interval = 0.5 if self.debug else 0.1
        while not teller.is_ready():
            if self.debug:
                print('waiting for ready')
            time.sleep(interval)
        self.get_ready_for_prepare()

    def get_ready_for_prepare(self):
        self.xr_data = self.teller.get_xr_data()
        self.folder_info = self.teller.get_folder_info()

        if self.for_sec:
            self.uv_data = self.teller.get_uv_data()
        else:
            self.uv_data = None

    def prepare(self):
        if self.teller is not None:
            self.wait_for_ready()
        try:
            if self.for_sec:
                self.prepare_for_sec()
            else:
                self.prepare_for_mf()
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            self.logger.error(str(etb))

    def prepare_for_sec(self):
        from SerialData import SerialData
        self.logger.info('preparing input for size-exclusive analysis.')

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            self.xr_data.plot()
            plt.show()

            self.uv_data.plot()
            plt.show()

        folder_info = self.folder_info

        print('folder_info.num_files=', folder_info.num_files)
        files = folder_info.get_files()

        xray_array = self.make_redundant_xray_data()
        uv_array = self.uv_data.data
        lvector = self.uv_data.vector
        col_header = None       # implement if needed
        mtd_elution = None
        data_info = [files, xray_array, uv_array, lvector, col_header, mtd_elution]
        in_folder = self.in_folder
        self.sd = SerialData(in_folder, in_folder, conc_file=folder_info.conc_file, data_info=data_info)
        self.files = files
        self.comments = folder_info.get_comments()

    def make_redundant_xray_data(self):
        # this is for backward compatibility and will be removed in the future
        return self.teller.get_redundant_xr_data()

    def prepare_for_mf(self):
        self.logger.info('preparing input for microfluidic analysis.')
        if self.teller is not None:
            xr_data = self.teller.get_xr_data()
            # print('xr_data.data.shape=', xr_data.data.shape)
            self.xr_data = xr_data

    def load_data(self, in_folder):
        from InputProcess.InputManager import InputManager
        from XrayData import XrayData
        from molass_legacy.UVData import UvData

        manager = InputManager(logger=self.logger)
        array_list = manager.load_data(in_folder, self.for_sec, self.xray_only, shared=False)
        self.xr_data = XrayData(None, array=array_list[0])
        self.folder_info = manager.get_folder_info()
        if self.for_sec:
            self.uv_data = UvData(in_folder, array=array_list[1], vector=array_list[2])
        else:
            self.uv_data = None
